
/* netndis.c - Linux Wireless LAN interface to NDIS wrapper
 *
 * Copyright (c) 2003 Linuxant inc.
 *
 * Based on orinoco.c
 *
 * (C) Copyright David Gibson, IBM Corporation 2001-2003.
 * Copyright (C) 2000 David Gibson, Linuxcare Australia <hhh@gibson.dropbear.id.au>
 *	With some help from :
 * Copyright (C) 2001 Jean Tourrilhes, HP Labs <jt@hpl.hp.com>
 * Copyright (C) 2001 Benjamin Herrenschmidt <benh@kernel.crashing.org>
 *
 * Based on dummy_cs.c 1.27 2000/06/12 21:27:25
 *
 * Portions based on wvlan_cs.c 1.0.6, Copyright Andreas Neuhaus <andy@fasta.fh-dortmund.de>
 *      http://www.stud.fh-dortmund.de/~andy/wvlan/
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * The initial developer of the original code is David A. Hinds
 * <dahinds@users.sourceforge.net>.  Portions created by David
 * A. Hinds are Copyright (C) 1999 David A. Hinds.  All Rights
 * Reserved.
 *
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License version 2 (the "GPL"), in
 * which case the provisions of the GPL are applicable instead of the
 * above.  If you wish to allow the use of your version of this file
 * only under the terms of the GPL and not to allow others to use your
 * version of this file under the MPL, indicate your decision by
 * deleting the provisions above and replace them with the notice and
 * other provisions required by the GPL.  If you do not delete the
 * provisions above, a recipient may use your version of this file
 * under either the MPL or the GPL.  */

/*
 * TODO
 *	o New wireless extensions API (patch from Moustafa
 *	  Youssef, updated by Jim Carter and Pavel Roskin).
 *	o Handle de-encapsulation within network layer, provide 802.11
 *	  headers (patch from Thomas 'Dent' Mirlacher)
 *	o RF monitor mode support
 *	o Fix possible races in SPY handling.
 *	o Disconnect wireless extensions from fundamental configuration.
 *	o (maybe) Software WEP support (patch from Stano Meduna).
 *	o (maybe) Use multiple Tx buffers - driver handling queue
 *	  rather than firmware. */

/* Locking and synchronization:
 *
 * The basic principle is that everything is serialized through a
 * single spinlock, priv->lock.  The lock is used in user, bh and irq
 * context, so when taken outside hardirq context it should always be
 * taken with interrupts disabled.  The lock protects both the
 * hardware and the struct netndis_private.
 *
 * Another flag, priv->hw_unavailable indicates that the hardware is
 * unavailable for an extended period of time (e.g. suspended, or in
 * the middle of a hard reset).  This flag is protected by the
 * spinlock.  All code which touches the hardware should check the
 * flag after taking the lock, and if it is set, give up on whatever
 * they are doing and drop the lock again.  The netndis_lock()
 * function handles this (it unlocks and returns -EBUSY if
 * hw_unavailable is non-zero). */


#define OS_NOSTRINGREDEFS

#include "oscompat.h"
#include "osservices.h"
#include "ndiswrap.h"

#include <linux/proc_fs.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/etherdevice.h>
#include <linux/wireless.h>
#if WIRELESS_EXT > 12
#include <net/iw_handler.h> // New driver API
#endif  /* WIRELESS_EXT > 12 */
#include <linux/ethtool.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/system.h>

#include "netndis.h"
#include "ieee802_11.h"

/********************************************************************/
/* Module information                                               */
/********************************************************************/

/* Level of debugging. Used in the macros in netndis.h */
#ifdef NETNDIS_DEBUG
static int netndis_debug = NETNDIS_DEBUG;
//MODULE_PARM(netndis_debug, "i");
//EXPORT_SYMBOL(netndis_debug);
#endif

static int suppress_linkstatus; /* = 0 */
MODULE_PARM(suppress_linkstatus, "i");

/********************************************************************/
/* Compile time configuration and compatibility stuff               */
/********************************************************************/

/* Wireless extensions backwards compatibility */
#ifndef SIOCIWFIRSTPRIV
#define SIOCIWFIRSTPRIV		SIOCDEVPRIVATE
#endif /* SIOCIWFIRSTPRIV */

/* We do this this way to avoid ifdefs in the actual code */
#ifdef WIRELESS_SPY
#define SPY_NUMBER(priv)	(priv->spy_number)
#else
#define SPY_NUMBER(priv)	0
#endif /* WIRELESS_SPY */

/********************************************************************/
/* Internal constants                                               */
/********************************************************************/

#define NETNDIS_MIN_MTU		256
#define NETNDIS_MAX_MTU		(IEEE802_11_DATA_LEN - ENCAPS_OVERHEAD)

#define SMALL_KEY_SIZE		5
#define LARGE_KEY_SIZE		13

enum {
	NETNDIS_WORK_RESET,
	NETNDIS_WORK_CONFIG,
};

/********************************************************************/
/* Data tables                                                      */
/********************************************************************/

/* The frequency of each channel in MHz (XXX: UPDATE) */
const long channel_frequency[] = {
	2412, 2417, 2422, 2427, 2432, 2437, 2442,
	2447, 2452, 2457, 2462, 2467, 2472, 2484
};
#define NUM_CHANNELS ( sizeof(channel_frequency) / sizeof(channel_frequency[0]) )

static struct proc_dir_entry *netndis_proc_dir;

static int __netndis_set_multicast_list(struct net_device *dev);
static void netndis_reset(struct net_device *dev);
static void netndis_config(struct net_device *dev);

static int netndis_open(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err;

	TRACE_ENTER(dev->name);

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	err = __netndis_up(dev);

	if (!err)
		priv->open = 1;

	netndis_unlock(priv);

	return err;
}

int netndis_stop(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err = 0, i;

	TRACE_ENTER(dev->name);

	for(i=0; (i<5) && priv->work_scheduled; i++) {
		OsTestPrintf("%s: waiting.. irql=%s\n", __FUNCTION__, OsIrqlName(OsGetIrql()));
		OsSleep(100);
	}

	/* We need to be able to close the interface even if hw_unavailable
	 * is set (e.g. as we're released after a PC Card removal) */
	err = netndis_lock(priv, TRUE);
	if (err)
		return err;

	priv->open = 0;

	err = __netndis_down(dev);

	netndis_unlock(priv);

	for(i=0; (i<5) && priv->work_scheduled; i++) {
		OsTestPrintf("%s: waiting.. irql=%s\n", __FUNCTION__, OsIrqlName(OsGetIrql()));
		OsSleep(100);
	}

	return err;
}

enum {
	NETNDIS_CFG_MAC,
	NETNDIS_CFG_MODE,
	NETNDIS_CFG_FREQ,
	NETNDIS_CFG_ESSID,
	NETNDIS_CFG_NICK,
	NETNDIS_CFG_RTS,
	NETNDIS_CFG_FRAG,
	NETNDIS_CFG_RATE,
	NETNDIS_CFG_PM,
	NETNDIS_CFG_ENC,
	NETNDIS_CFG_KEYS,
	NETNDIS_CFG_AUTHMODE,
	NETNDIS_CFG_MULTI,
	NETNDIS_CFG_LAST
};

static int __netndis_hw_set_bitrate(struct netndis_private *priv, struct iw_param *bitratemode)
{
	int err;
#ifndef DLDR_DESIRED_RATES_OID_SUPPORTED
	unsigned char supportedrates[8];
#if 0
	ULONG UVal;

	err = NdisWrapStatusToErrno(
			NdisWrapGetLinkSpeed(priv->miniportAdapter, &UVal));
	if (err)
		return err;

	if(UVal * 100 != bitratemode->value) {
		err = -EINVAL;
	}
#endif

	supportedrates[0] = (bitratemode->value == -1) ? 0 : (bitratemode->value / 500000);
	memset(&supportedrates[1], 0, sizeof(supportedrates) - 1);

	err = NdisWrapStatusToErrno(
			NdisWrapSetDesiredRates(priv->miniportAdapter, supportedrates, sizeof(supportedrates)));

#else
	int i;
	unsigned char supportedrates[16];
	int nrates;

	err = NdisWrapStatusToErrno(
			NdisWrapGetSupportedRates(priv->miniportAdapter, supportedrates, sizeof(supportedrates)));
	if (err)
		return err;

	for(nrates=0; supportedrates[nrates] && (nrates < sizeof(supportedrates)); nrates++);

	/* As the user space doesn't know our highest rate, it uses -1
	 * to ask us to set the highest rate.  Test it using "iwconfig
	 * ethX rate auto" - Jean II */
	if (bitratemode->value == -1) {
		err = NdisWrapStatusToErrno(
			NdisWrapSetDesiredRates(priv->miniportAdapter, supportedrates, nrates));
	} else {
		for(i=0; i < nrates; i++) {
			if((supportedrates[i] & 0x7f) * 500000 == bitratemode->value) {
				if(bitratemode->fixed) {
					supportedrates[0] = supportedrates[i];
					i = 0;
				}
				memset(&supportedrates[i+1], 0, sizeof(supportedrates) - (i+1));
				/* 8 is original native NDIS size for this request */
				nrates = (i < 8) ? 8 : (i + 1);

				err = NdisWrapStatusToErrno(
						NdisWrapSetDesiredRates(priv->miniportAdapter, supportedrates, nrates));
			}
		}

		if(i == nrates) {
			err = -EINVAL; /* rate not found */
		}
	}
#endif

	if((err == -EOPNOTSUPP) || (err == -ENOSYS) || (err == -EINVAL))
		err = 0; /* silently ignore for now to accomodate redhat-config-network */

	return err;
}

static int __netndis_hw_set_authmode(struct netndis_private *priv)
{
	int err;
	BOOLEAN wep_on = !(priv->iw_encode & IW_ENCODE_DISABLED);
	ndiswrapauthmode_t authmode;

again:
	if(!wep_on || (priv->iw_encode & IW_ENCODE_OPEN))
		authmode = NDISWRAP_AUTHMODE_OPEN;
	else if(priv->iw_encode & IW_ENCODE_RESTRICTED)
		authmode = NDISWRAP_AUTHMODE_SHARED;
	else
		authmode = NDISWRAP_AUTHMODE_AUTOSWITCH;

	err = NdisWrapStatusToErrno(
			NdisWrapSetAuthMode(priv->miniportAdapter, authmode));
	if(err && authmode != NDISWRAP_AUTHMODE_OPEN) {
		/* XXX is this proper? */
		priv->iw_encode &= ~IW_ENCODE_RESTRICTED;
		priv->iw_encode |= IW_ENCODE_OPEN;
		goto again;
	}

	return err;
}

static int __netndis_hw_set_keys(struct netndis_private *priv)
{
	int err = 0, err2;
	BOOLEAN wep_on = !(priv->iw_encode & IW_ENCODE_DISABLED);

	if(wep_on) {
		int i;

#if 0
		err2 = NdisWrapStatusToErrno(
				NdisWrapWepReloadDefaults(priv->miniportAdapter));
#else
		err2 = -EOPNOTSUPP; /* XXX */
#endif

		/* Write all 4 keys */
		for(i = 0; i < NETNDIS_MAX_KEYS; i++) {

			if(!err2 && !priv->keys[i].len)
				continue;

			if (priv->keys[i].len	> LARGE_KEY_SIZE) {
				printk(KERN_ERR "%s: BUG: Key %d has oversize length %d.\n",
						priv->ndev->name, i, priv->keys[i].len);
				return -E2BIG;
			}

			err = NdisWrapStatusToErrno(
					NdisWrapSetWepKey(priv->miniportAdapter, i, priv->keys[i].data, priv->keys[i].len, i == priv->tx_key));
			if (err && priv->keys[i].len)
				return err;
		}
	}

	return err;
}

static int __netndis_hw_set_enc(struct netndis_private *priv)
{
	int err;
	BOOLEAN wep_on = !(priv->iw_encode & IW_ENCODE_DISABLED);

	err = NdisWrapStatusToErrno(
			NdisWrapSetWep(priv->miniportAdapter, wep_on));
	if(err)
		return err;

#if 0
	/*err2 =*/ NdisWrapStatusToErrno(
			NdisWrapSetPrivacyFilter(priv->miniportAdapter, wep_on));
#endif

	NdisWrapClearCachedSSID(priv->miniportAdapter);

	return err;
}

static int __netndis_configure(struct net_device *dev, BOOLEAN enablereceive)
{
	struct netndis_private *priv = dev->priv;
	int err;

#if 0
	err = NdisWrapStatusToErrno(
			NdisWrapEnableReceive(priv->miniportAdapter, FALSE, FALSE, FALSE, FALSE, FALSE));
	if (err) {
		printk(KERN_WARNING "%s: Unable to disable port to configure card\n",
		       dev->name);
		return err;
	}
#endif

	/* Get the MAC address */
	if(test_and_clear_bit(NETNDIS_CFG_MAC, &priv->cfg)) {
		unsigned char current_devaddr[ETH_ALEN];

		err = NdisWrapStatusToErrno(
				NdisWrapGetCurrentEthAddress(priv->miniportAdapter, current_devaddr, ETH_ALEN));
		if (err) {
			printk(KERN_WARNING "%s: failed to read MAC address!\n",
					dev->name);
			return err;
		}

		/* Set the MAC address, if different */
		if(memcmp(current_devaddr, dev->dev_addr, ETH_ALEN)) {
			err = NdisWrapStatusToErrno(
					NdisWrapSetCurrentEthAddress(priv->miniportAdapter, dev->dev_addr, ETH_ALEN));
			if (err) {
				printk(KERN_ERR "%s: Error %d setting MAC address\n", dev->name, err);
				return err;
			}
		}
	}

	/* Set the channel/frequency */
	if(test_and_clear_bit(NETNDIS_CFG_FREQ, &priv->cfg)) {
#if 0
		if (priv->channel == 0) {
			printk(KERN_DEBUG "%s: Channel is 0 in __netndis_configure()\n", dev->name);
			if (priv->createibss)
				priv->channel = 10;
		}
		err = hhh_write_wordrec(hw, USER_BAP, HHH_RID_CNFOWNCHANNEL, priv->channel);
		if (err) {
			printk(KERN_ERR "%s: Error %d setting channel\n", dev->name, err);
			return err;
		}
#endif
	}

	/* Set the station name */
	if(test_and_clear_bit(NETNDIS_CFG_NICK, &priv->cfg)) {
		err = NdisWrapStatusToErrno(
				NdisWrapSetMachineName(priv->miniportAdapter, priv->nick));
		if (err && err != -ENOSYS && err != -EOPNOTSUPP) {
			printk(KERN_ERR "%s: Error %d setting nickname\n", dev->name, err);
			return err;
		}
	}

#if 0
	/* Set AP density */
	if (priv->has_sensitivity) {
		err = hhh_write_wordrec(hw, USER_BAP, HHH_RID_CNFSYSTEMSCALE,
					   priv->ap_density);
		if (err) {
			printk(KERN_WARNING "%s: Error %d setting SYSTEMSCALE.  "
			       "Disabling sensitivity control\n", dev->name, err);

			priv->has_sensitivity = 0;
		}
	}
#endif

	/* Set RTS threshold */
	if(test_and_clear_bit(NETNDIS_CFG_RTS, &priv->cfg)) {
		err = NdisWrapStatusToErrno(
				NdisWrapSetRtsThreshHold(priv->miniportAdapter, priv->rts_thresh));
		if (err) {
			//printk(KERN_ERR "%s: Error %d setting RTS threshold\n", dev->name, err);
		}
	}

	/* Set fragmentation threshold */
	if(test_and_clear_bit(NETNDIS_CFG_FRAG, &priv->cfg)) {
		err = NdisWrapStatusToErrno(
				NdisWrapSetFragThreshHold(priv->miniportAdapter, priv->frag_thresh));
		if (err) {
			//printk(KERN_ERR "%s: Error %d setting framentation\n", dev->name, err);
		}
	}

	/* Set bitrate */
	if(test_and_clear_bit(NETNDIS_CFG_RATE, &priv->cfg)) {
		if(priv->bitratemode.value) {
			err = __netndis_hw_set_bitrate(priv, &priv->bitratemode);
			if (err) {
				printk(KERN_ERR "%s: Error %d setting bitrate\n", dev->name, err);
				return err;
			}
		}
	}

	/* Set power management */
	if(test_and_clear_bit(NETNDIS_CFG_PM, &priv->cfg) && priv->has_pm) {
		ndiswrappowermode_t powermode = priv->pm_on ? NDISWRAP_PWRMODE_FAST_PSP : NDISWRAP_PWRMODE_CAM;

		err = NdisWrapStatusToErrno(
				NdisWrapSetPowerMode(priv->miniportAdapter, powermode));
		if (err) {
			printk(KERN_ERR "%s: Error %d setting up PM\n",
			       dev->name, err);
			return err;
		}

#if 0
		err = hhh_write_wordrec(hw, USER_BAP,
					   HHH_RID_CNFMAXSLEEPDURATION,
					   priv->pm_period);
		if (err) {
			printk(KERN_ERR "%s: Error %d setting up PM\n",
			       dev->name, err);
			return err;
		}
		err = hhh_write_wordrec(hw, USER_BAP,
					   HHH_RID_CNFPMHOLDOVERDURATION,
					   priv->pm_timeout);
		if (err) {
			printk(KERN_ERR "%s: Error %d setting up PM\n",
			       dev->name, err);
			return err;
		}
#endif
	}

	/* Set up the link mode */
	if(test_and_clear_bit(NETNDIS_CFG_MODE, &priv->cfg)) {
		err = NdisWrapStatusToErrno(
				NdisWrapSetInfraMode(priv->miniportAdapter, priv->iw_mode));
		if (err) {
			printk(KERN_ERR "%s: Error %d setting link mode\n", dev->name, err);
			return err;
		}
	}

	/* Set up encryption */
	if(test_and_clear_bit(NETNDIS_CFG_AUTHMODE, &priv->cfg) && priv->has_wep) {
		err = __netndis_hw_set_authmode(priv);
		if (err) {
			printk(KERN_ERR "%s: Error %d setting authentication mode\n",
			       dev->name, err);
			priv->has_wep = 0;
		}
	}
	if(test_and_clear_bit(NETNDIS_CFG_KEYS, &priv->cfg) && priv->has_wep) {
		err = __netndis_hw_set_keys(priv);
		if(err == -ENETDOWN) { // adapter not ready?
			set_bit(NETNDIS_CFG_KEYS, &priv->cfg);
			err = 0;
		}
		if (err) {
			printk(KERN_ERR "%s: Error %d configuring keys\n",
			       dev->name, err);
			//priv->has_wep = 0;
			//err=0;
		}
	}
	if(test_and_clear_bit(NETNDIS_CFG_ENC, &priv->cfg) && priv->has_wep) {
		err = __netndis_hw_set_enc(priv);
		if (err) {
			printk(KERN_ERR "%s: Error %d activating encryption\n",
			       dev->name, err);
			priv->has_wep = 0;
		}
	}

	/* Set promiscuity / multicast */
	if(test_and_clear_bit(NETNDIS_CFG_MULTI, &priv->cfg)) {
		err = __netndis_set_multicast_list(dev); /* FIXME: what about the xmit_lock */
		if (err) {
			err = 0; /* XXX ignore for now */
		}
	}

	if(enablereceive) {
		BOOLEAN EnableReceive = priv->open && netif_carrier_ok(dev);
again:
		err = NdisWrapStatusToErrno(
				NdisWrapEnableReceive(priv->miniportAdapter, EnableReceive, EnableReceive, EnableReceive && priv->pm_mcast && priv->mc_count, EnableReceive && priv->allmulti, priv->promiscuous));
		if (err) {
			if(err == -EOPNOTSUPP && priv->promiscuous) {
				priv->promiscuous = 0;
				dev->flags &= ~IFF_PROMISC;
				goto again;
			}
			if(err == -ECONNREFUSED) {
				if (time_before(priv->print_timestamp, jiffies)) {
					printk(KERN_ERR "%s: License not valid\n", dev->name);
					priv->print_timestamp = jiffies + 5 * HZ;
				}
			} else
				printk(KERN_ERR "%s: Error %d enabling MAC port\n",
						dev->name, err);
			return err;
		}
#if 0
		/* Since we can set the promiscuous flag when it wasn't asked
		   for, make sure the net_device knows about it. */
		if (priv->promiscuous)
			dev->flags |= IFF_PROMISC;
		else
			dev->flags &= ~IFF_PROMISC;
#endif
	}

	/* Set the desired ESSID */
	if(test_and_clear_bit(NETNDIS_CFG_ESSID, &priv->cfg)) {
		err = NdisWrapStatusToErrno(
				NdisWrapSetSSID(priv->miniportAdapter, priv->desired_essid));
		if (err) {
			printk(KERN_ERR "%s: Error %d setting desired SSID\n", dev->name, err);
			return err;
		}
	}

	return 0;
}

int __netndis_up(struct net_device *dev)
{
	//struct netndis_private *priv = dev->priv;
	int err;

	OsDebugPrintf("%s\n", __FUNCTION__);

#if 0
	netif_carrier_off(dev); /* just to make sure */
#endif

	err = __netndis_configure(dev, TRUE);
	if (err) {
		printk(KERN_ERR "%s: Error %d configuring card\n",
		       dev->name, err);
		return err;
	}

	netif_start_queue(dev);

	return 0;
}

int __netndis_down(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err;

	OsDebugPrintf("%s\n", __FUNCTION__);
	netif_stop_queue(dev);

	if (!priv->hw_unavailable) {
		err = NdisWrapStatusToErrno(
				NdisWrapEnableReceive(priv->miniportAdapter, FALSE, FALSE, FALSE, FALSE, FALSE));
		if (err) {
			printk(KERN_WARNING "%s: Error %d disabling MAC port\n",
					dev->name, err);
		}
	}
	
#if 0
	/* firmware will have to reassociate */
	netif_carrier_off(dev);
	priv->last_linkstatus = NDISWRAP_LINKSTATUS_LAST;
#endif

	return 0;
}

/* This must be called from user context, without locks held - use
 * schedule_work() */
static void netndis_config(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err;

	OsDebugPrintf("%s\n", __FUNCTION__);

	err = netndis_lock(priv, FALSE);
	if (err) {
		/* When the hardware becomes available again, whatever
		 * detects that is responsible for re-initializing
		 * it. So no need for anything further*/
		OsModuleUseCountDec();
		clear_bit(NETNDIS_WORK_CONFIG, &priv->work_scheduled);
		return;
	}

#if 0
	__netndis_down(dev);
	if (priv->open) {
		err = __netndis_up(dev);
	}
#else
	err = __netndis_configure(dev, TRUE);
	if (err) {
		printk(KERN_ERR "%s: Error %d configuring card\n",
		       dev->name, err);
		OsModuleUseCountDec();
		clear_bit(NETNDIS_WORK_CONFIG, &priv->work_scheduled);
		return;
	}
#endif

	netndis_unlock(priv);

	netif_wake_queue(dev);

	OsModuleUseCountDec();
	clear_bit(NETNDIS_WORK_CONFIG, &priv->work_scheduled);
}


/********************************************************************/
/* Internal helper functions                                        */
/********************************************************************/

/* Does the frame have a SNAP header indicating it should be
 * de-encapsulated to Ethernet-II? */
static inline int
is_ethersnap(struct header_struct *hdr)
{
	/* We de-encapsulate all packets which, a) have SNAP headers
	 * (i.e. SSAP=DSAP=0xaa and CTRL=0x3 in the 802.2 LLC header
	 * and where b) the OUI of the SNAP header is 00:00:00 or
	 * 00:00:f8 - we need both because different APs appear to use
	 * different OUIs for some reason */
	return (memcmp(&hdr->dsap, &encaps_hdr, 5) == 0)
		&& ( (hdr->oui[2] == 0x00) || (hdr->oui[2] == 0xf8) );
}

static void
netndis_set_multicast_list(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;

	OsDebugPrintf("%s\n", __FUNCTION__);

	set_bit(NETNDIS_CFG_MULTI, &priv->cfg);

	if (!priv->hw_unavailable) {
		if(!test_and_set_bit(NETNDIS_WORK_CONFIG, &priv->work_scheduled)) {
			OsDebugPrintf("%s: scheduled config_work\n", __FUNCTION__);
			if(OsThreadSchedule(priv->osthrd, &priv->config_work, (void (*)(void *))netndis_config, dev) <= 0) {
				clear_bit(NETNDIS_WORK_CONFIG, &priv->work_scheduled);
			}
		}
	}
}


static int netndis_hw_get_bssid(struct netndis_private *priv,
				char buf[ETH_ALEN])
{
	int err = 0;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	err = NdisWrapStatusToErrno(
			NdisWrapGetBSSID(priv->miniportAdapter, buf, ETH_ALEN));

	if(err == -ENETDOWN) {
		memset(buf, 0, ETH_ALEN);
		err = 0;
	} else if(err) {
		printk(KERN_ERR"%s: NdisWrapGetBSSID returned err=%d\n", __FUNCTION__, err);
	}

	netndis_unlock(priv);

	return err;
}

static int netndis_hw_get_essid(struct netndis_private *priv, int *active,
				char buf[IW_ESSID_MAX_SIZE+1])
{
	int err = 0;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	if (strlen(priv->desired_essid) > 0) {
		*active = 1;

	} else {
		*active = 0;

	}

	memset(buf, 0, IW_ESSID_MAX_SIZE+1);

	err = NdisWrapStatusToErrno(
			NdisWrapGetSSID(priv->miniportAdapter, buf, IW_ESSID_MAX_SIZE+1));
	if (err) {
		if(err == -ENETDOWN) {
			err = 0;
		} else {
			printk(KERN_ERR"%s: NdisWrapGetSSID returned err=%d\n", __FUNCTION__, err);
		}
	}

	netndis_unlock(priv);

	return err;       
}

static long netndis_hw_get_freq(struct netndis_private *priv)
{
	
	int err = 0;
	u16 channel;
	long freq = 0;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	
#if 0
	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_CURRENTCHANNEL, &channel);
#else
	err = -EOPNOTSUPP; /* XXX */
#endif
	if (err)
		goto out;

	/* Intersil firmware 1.3.5 returns 0 when the interface is down */
	if (channel == 0) {
		err = -EBUSY;
		goto out;
	}

	if ( (channel < 1) || (channel > NUM_CHANNELS) ) {
		printk(KERN_WARNING "%s: Channel out of range (%d)!\n",
		       priv->ndev->name, channel);
		err = -EBUSY;
		goto out;

	}
	freq = channel_frequency[channel-1] * 100000;

 out:
	netndis_unlock(priv);

	if (err > 0)
		err = -EBUSY;
	return err ? err : freq;
}

static int netndis_hw_get_bitratelist(struct netndis_private *priv,
				      int *numrates, s32 *rates, int max)
{
	int err = 0;
	int i, nrates;
	unsigned char supportedrates[16], *p;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	err = NdisWrapStatusToErrno(
			NdisWrapGetSupportedRates(priv->miniportAdapter, supportedrates, sizeof(supportedrates)));

	netndis_unlock(priv);

	if (err)
		return err;
	
	for(nrates=0; supportedrates[nrates] && (nrates < sizeof(supportedrates)); nrates++);

	if(max < nrates) {
		p = &supportedrates[nrates - max];
		nrates = max;
	} else
		p = supportedrates;

	for (i = 0; i < nrates; i++) {
		rates[i] = (p[i] & 0x7f) * 500000; /* convert to bps */
	}

	*numrates = nrates;

	return 0;
}

static void print_linkstatus(struct net_device *dev, ndiswraplinkstatus_t status)
{
	char * s;

	if (suppress_linkstatus)
		return;

	switch (status) {
	case NDISWRAP_LINKSTATUS_NOT_CONNECTED:
		s = "Not Connected";
		break;
	case NDISWRAP_LINKSTATUS_CONNECTED:
		s = "Connected";
		break;
	case NDISWRAP_LINKSTATUS_DISCONNECTED:
		s = "Disconnected";
		break;
	case NDISWRAP_LINKSTATUS_AP_CHANGE:
		s = "AP Changed";
		break;
	case NDISWRAP_LINKSTATUS_AP_OUT_OF_RANGE:
		s = "AP Out of Range";
		break;
	case NDISWRAP_LINKSTATUS_AP_IN_RANGE:
		s = "AP In Range";
		break;
	case NDISWRAP_LINKSTATUS_ASSOC_FAILED:
		s = "Association Failed";
		break;
	default:
		s = "UNKNOWN";
	}
	
	printk(KERN_INFO "%s: New link status: %s (%04x)\n",
	       dev->name, s, status);
}

__shimcall__
static void
netndis_status_handler(struct net_device *dev, ndiswraplinkstatus_t newstatus)
{
	struct netndis_private *priv = ((struct net_device *)dev)->priv;

	if ( (newstatus == NDISWRAP_LINKSTATUS_CONNECTED)
			|| (newstatus == NDISWRAP_LINKSTATUS_AP_CHANGE)
			|| (newstatus == NDISWRAP_LINKSTATUS_AP_IN_RANGE) ) {
		netif_carrier_off(dev);
		netif_carrier_on(dev);
	} else if ( (newstatus == NDISWRAP_LINKSTATUS_NOT_CONNECTED)
			|| (newstatus == NDISWRAP_LINKSTATUS_DISCONNECTED)
			|| (newstatus == NDISWRAP_LINKSTATUS_AP_OUT_OF_RANGE)
			|| (newstatus == NDISWRAP_LINKSTATUS_ASSOC_FAILED) )
		netif_carrier_off(dev);

	if (newstatus != priv->last_linkstatus) {
		print_linkstatus(dev, newstatus);
		set_bit(NETNDIS_CFG_KEYS, &priv->cfg);
	}

	priv->last_linkstatus = newstatus;

	if (!priv->hw_unavailable) {
		if(!test_and_set_bit(NETNDIS_WORK_CONFIG, &priv->work_scheduled)) {
			OsDebugPrintf("%s: scheduled config_work\n", __FUNCTION__);
			if(OsThreadSchedule(priv->osthrd, &priv->config_work, (void (*)(void *))netndis_config, dev) <= 0) {
				clear_bit(NETNDIS_WORK_CONFIG, &priv->work_scheduled);
			}
		}
	}
}

#ifdef NETNDIS_ZEROCOPY
static void
netndis_return_packets(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int i;

	TRACE_ENTER(dev->name);

	if(netndis_lock(priv, FALSE)) {
		printk(KERN_ERR"%s: lock failed\n", __FUNCTION__);
		return;
	}

	for(i = 0; i < NETNDIS_NUM_PCKTCTX; i++) {
		if(priv->packetctxarray[i].page) {
			if(priv->packetctxarray[i].initialcount == atomic_read(&priv->packetctxarray[i].page->count)) {
				put_page(priv->packetctxarray[i].page);
				NdisWrapReturnPacket(priv->packetctxarray[i].packetctx);
				priv->packetctxarray[i].page = NULL;
			}
		}
	}

	netndis_unlock(priv);

	TRACE_EXIT(dev->name);
}
#endif

__shimcall__
static void *
netndis_rx_handler(struct net_device *dev, int status,
	void *packetctx, unsigned int headersize, void *buffer, unsigned int length)
{
	struct netndis_private *priv = dev->priv;
	struct net_device_stats *stats = &priv->stats;
	struct sk_buff *skb = NULL;
	void *p;
	int i;
#ifdef NETNDIS_ZEROCOPY
	unsigned int l;
	static BOOLEAN zerocopy_disabled;
	BOOLEAN zerocopy = (!zerocopy_disabled) && (!status);
#endif

	if((headersize != ETH_HLEN) || (length < ETH_HLEN)) {
		goto drop;
	}

#define NETNDIS_SKBRES_LEN 2

#ifdef NETNDIS_ZEROCOPY
tryagain:
#define NETNDIS_ZEROCOPY_HDRLEN (headersize + 16)
	skb = dev_alloc_skb(NETNDIS_SKBRES_LEN + (zerocopy ? NETNDIS_ZEROCOPY_HDRLEN : length));
#else
	skb = dev_alloc_skb(NETNDIS_SKBRES_LEN + length);
#endif
	if (!skb) {
		printk(KERN_WARNING "%s: Can't allocate skb for Rx\n", dev->name);
		goto drop;
	}

	skb_reserve(skb, NETNDIS_SKBRES_LEN); /* This way the IP header is aligned */

#ifdef NETNDIS_ZEROCOPY
	if(zerocopy) {
		if(length < NETNDIS_ZEROCOPY_HDRLEN)
			l = length;
		else
			l = NETNDIS_ZEROCOPY_HDRLEN;

		p = skb_put(skb, l);
		memcpy(p, buffer, l);

		buffer += l;
		length -= l;

		p = buffer; l = length;

		for(i=0; l && (i < MAX_SKB_FRAGS); i++) {
			__u16 fragsize;

			skb_shinfo(skb)->frags[i].page = virt_to_page(p);
			get_page(skb_shinfo(skb)->frags[i].page); /* to compensate for the put_page() in skb_release_data() */
#ifdef CONFIG_HIGHMEM
			if(skb_shinfo(skb)->frags[i].page >= highmem_start_page) {
				printk(KERN_WARNING "%s: disabling zerocopy (highmem not supported)\n", dev->name);
				zerocopy_disabled = TRUE;
				zerocopy = FALSE;
				dev_kfree_skb_irq(skb);
				goto tryagain;
			}
#endif

			skb_shinfo(skb)->frags[i].page_offset = p - page_address(skb_shinfo(skb)->frags[i].page);

			fragsize = (l < PAGE_SIZE ? l : PAGE_SIZE);

			skb_shinfo(skb)->frags[i].size = fragsize;

			l -= fragsize;
			p += fragsize;

			skb->len += fragsize;
			skb->data_len += fragsize;
		}

		skb_shinfo(skb)->nr_frags = i;

		if(skb_shinfo(skb)->nr_frags) {
			if(netndis_lock(priv, FALSE))
				goto drop;

			for(i = 0; i < NETNDIS_NUM_PCKTCTX; i++) {
				if(priv->packetctxarray[i].page) {
					if(priv->packetctxarray[i].initialcount == atomic_read(&priv->packetctxarray[i].page->count)) {
						put_page(priv->packetctxarray[i].page);
						NdisWrapReturnPacket(priv->packetctxarray[i].packetctx);
						priv->packetctxarray[i].page = NULL;
					}
				}
				if(!priv->packetctxarray[i].page) {
					priv->packetctxarray[i].packetctx = packetctx;
					priv->packetctxarray[i].initialcount = atomic_read(&skb_shinfo(skb)->frags[0].page->count);
					get_page(skb_shinfo(skb)->frags[0].page);
					priv->packetctxarray[i].page = skb_shinfo(skb)->frags[0].page;
					break;
				}
			}
			netndis_unlock(priv);

			if(i == NETNDIS_NUM_PCKTCTX) {
				printk(KERN_WARNING "%s: packetctxarray full\n", dev->name);
				zerocopy_disabled = TRUE;
				zerocopy = FALSE;
				dev_kfree_skb_irq(skb);
				goto tryagain;
			}
		}

	} else
#endif
	{
		p = skb_put(skb, length);
		memcpy(p, buffer, length);
	}

	dev->last_rx = jiffies;
	skb->dev = dev;
	skb->protocol = eth_type_trans(skb, dev);
	skb->ip_summed = CHECKSUM_NONE;

	/* Using spy support with lots of Rx packets, like in an
	 * infrastructure (AP), will really slow down everything, because
	 * the MAC address must be compared to each entry of the spy list.
	 * If the user really asks for it (set some address in the
	 * spy list), we do it, but he will pay the price.
	 * Note that to get here, you need both WIRELESS_SPY
	 * compiled in AND some addresses in the list !!!
	 */
	/* Note : gcc will optimise the whole section away if
	 * WIRELESS_SPY is not defined... - Jean II */
	if (SPY_NUMBER(priv)) {
		LONG Rssi, NoiseLevel;

		/* Gather wireless spy statistics: for each packet, compare the
		 * source address with out list, and if match, get the stats... */
		if(!NdisWrapStatusToErrno(
				NdisWrapGetRssi(priv->miniportAdapter, &Rssi))) {

		if(NdisWrapStatusToErrno(
				NdisWrapGetNoiseLevel(priv->miniportAdapter, &NoiseLevel)))
			NoiseLevel = -200;

			for (i = 0; i < priv->spy_number; i++)
				if (!memcmp(skb->mac.raw + ETH_ALEN, priv->spy_address[i], ETH_ALEN)) {
					priv->spy_stat[i].level = 0x100 + Rssi;
					priv->spy_stat[i].noise = 0x100 + NoiseLevel;
					priv->spy_stat[i].qual = 1;
					priv->spy_stat[i].updated = 3;
				}
		}
	}

	/* Pass the packet to the networking stack */
	netif_rx(skb);
	stats->rx_packets++;
	stats->rx_bytes += length;

#ifdef NETNDIS_ZEROCOPY
	netndis_return_packets(dev);
	if(zerocopy)
		return skb;
	else
#endif
		return NULL;

 drop:	
	stats->rx_dropped++;

	if (skb)
		dev_kfree_skb_irq(skb);
#ifdef NETNDIS_ZEROCOPY
	netndis_return_packets(dev);
#endif
	return NULL;
}

/*
 * struct net_device methods
 */

struct net_device_stats *
netndis_get_stats(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	
	return &priv->stats;
}

struct iw_statistics *
netndis_get_wireless_stats(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	struct iw_statistics *wstats = &priv->wstats;
	int err = 0;
	ndiswraplinkstats_t linkstats;

	if (!netif_device_present(dev)) {
		printk(KERN_WARNING "%s: get_wireless_stats() called while device not present\n",
		       dev->name);
		return NULL; /* FIXME: Can we do better than this? */
	}

	err = netndis_lock(priv, FALSE);
	if (err)
		return NULL; /* FIXME: Erg, we've been signalled, how
			      * do we propagate this back up? */

	NdisWrapStatusToErrno(
			NdisWrapGetLinkStats(priv->miniportAdapter, &linkstats));
	wstats->discard.retries = linkstats.Faileds;
	wstats->discard.misc = linkstats.RTSFailures + linkstats.ACKFailures + linkstats.FrameDuplicates + linkstats.FCSErrors;

	if (priv->iw_mode == IW_MODE_ADHOC) {
		memset(&wstats->qual, 0, sizeof(wstats->qual));
		/* If a spy address is defined, we report stats of the
		 * first spy address - Jean II */
		if (SPY_NUMBER(priv)) {
			wstats->qual.qual = priv->spy_stat[0].qual;
			wstats->qual.level = priv->spy_stat[0].level;
			wstats->qual.noise = priv->spy_stat[0].noise;
			wstats->qual.updated = priv->spy_stat[0].updated;
		}
	} else {
		LONG Rssi, NoiseLevel;

		if(!NdisWrapStatusToErrno(
				NdisWrapGetRssi(priv->miniportAdapter, &Rssi))) {

			if(NdisWrapStatusToErrno(
						NdisWrapGetNoiseLevel(priv->miniportAdapter, &NoiseLevel)))
				NoiseLevel = -200;

			wstats->qual.qual = 1;
			wstats->qual.level = 0x100 + Rssi;
			wstats->qual.noise = 0x100 + NoiseLevel;
			wstats->qual.updated = 2;
		}
	}

	netndis_unlock(priv);

	return err ? NULL : wstats;
}

__shimcall__
static void
netndis_xmitcomplete_handler(struct net_device *dev, int err, struct sk_buff *skb)
{
	struct netndis_private *priv = dev->priv;
	struct net_device_stats *stats = &priv->stats;

	dev_kfree_skb(skb);

	if(err) {
		printk(KERN_ERR"%s: err %d\n", __FUNCTION__, err);
		stats->tx_errors++;
	} else
		stats->tx_packets++;

	netif_wake_queue(dev);
}

static int
netndis_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct netndis_private *priv = (struct netndis_private *)dev->priv;
	struct net_device_stats *stats = &priv->stats;
	int err = 0;
	int len;

	TRACE_ENTER(dev->name);

	if (!netif_running(dev)) {
		printk(KERN_ERR "%s: Tx on stopped device!\n",
		       dev->name);
		TRACE_EXIT(dev->name);
		return 1;
	}
	
	if (netif_queue_stopped(dev)) {
		printk(KERN_DEBUG "%s: Tx while transmitter busy!\n", 
		       dev->name);
		TRACE_EXIT(dev->name);
		return 1;
	}
	
	if (netndis_lock(priv, FALSE) != 0) {
		//printk(KERN_ERR "%s: netndis_xmit() called while hw_unavailable\n", dev->name);
		TRACE_EXIT(dev->name);
/*  		BUG(); */
		return 1;
	}

	if (!netif_carrier_ok(dev)) {
		/* Oops, the firmware hasn't established a connection,
                   silently drop the packet (this seems to be the
                   safest approach). */
		stats->tx_errors++;
		netndis_unlock(priv);
		dev_kfree_skb(skb);
		TRACE_EXIT(dev->name);
		return 0;
	}

	netif_stop_queue(dev);

	len = skb->len;

	err = NdisWrapStatusToErrno(
			NdisWrapTransmit(priv->miniportAdapter, skb->data, skb->len, skb));
	if (err && err != -EWOULDBLOCK) {
		netif_start_queue(dev);
		printk(KERN_ERR "%s: Error %d transmitting packet\n",
		       dev->name, err);
		stats->tx_errors++;
		goto fail;
	}

	dev->trans_start = jiffies;
	stats->tx_bytes += len;

	netndis_unlock(priv);

	TRACE_EXIT(dev->name);

	return 0;
 fail:
	TRACE_EXIT(dev->name);

	netndis_unlock(priv);
	return err;
}

static void
netndis_tx_timeout(struct net_device *dev)
{
	struct netndis_private *priv = (struct netndis_private *)dev->priv;
	struct net_device_stats *stats = &priv->stats;

	//printk(KERN_WARNING "%s: Tx timeout! ", dev->name);

	stats->tx_errors++;

#if 0
	if(!test_and_set_bit(NETNDIS_WORK_RESET, &priv->work_scheduled)) {
		OsTestPrintf("%s: scheduled reset_work\n", __FUNCTION__);
		if(OsThreadSchedule(priv->osthrd, &priv->reset_work, (void (*)(void *))netndis_reset, dev) <= 0) {
			clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
		}
	}
#endif
}

static int
netndis_change_mtu(struct net_device *dev, int new_mtu)
{
	//struct netndis_private *priv = dev->priv;

	if ( (new_mtu < NETNDIS_MIN_MTU) || (new_mtu > NETNDIS_MAX_MTU) )
		return -EINVAL;

	if ( (new_mtu + ENCAPS_OVERHEAD + IEEE802_11_HLEN) >
	     ((IEEE802_11_FRAME_LEN + ETH_HLEN) - ETH_HLEN) )
		return -EINVAL;

	dev->mtu = new_mtu;

	return 0;
}

static int
__netndis_set_multicast_list(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	int mc_count;

	//printk("%s: dev=%p\n", __FUNCTION__, dev);

	if((dev->flags & IFF_ALLMULTI) || (dev->mc_count > priv->max_multicast)) {
		priv->allmulti = 1;
		mc_count = 0;
	} else {
		priv->allmulti = 0;
		mc_count = dev->mc_count;
	}

	if (dev->flags & IFF_PROMISC) {
		priv->promiscuous = 1;
	} else {
		priv->promiscuous = 0;
	}

	if (mc_count) {
		struct dev_mc_list *p = dev->mc_list;
		unsigned char *mclist;
		int i;

		mclist = OsAllocate(mc_count * ETH_ALEN);
		if(!mclist) {
			printk(KERN_ERR"%s: cannot allocate %d bytes for mclist\n", dev->name, mc_count * ETH_ALEN);
			err = -ENOMEM;
			mc_count = 0;
		} else {
			for (i = 0; i < mc_count; i++) {
				/* Paranoia: */
				if (!p)
					BUG(); /* Multicast list shorter than mc_count */
				if (p->dmi_addrlen != ETH_ALEN)
					BUG(); /* Bad address size in multicast list */

				memcpy(mclist + (i*ETH_ALEN), p->dmi_addr, ETH_ALEN);
				p = p->next;
			}

			if (p)
				printk(KERN_WARNING "Multicast list is longer than mc_count\n");

			err = NdisWrapStatusToErrno(
					NdisWrapSetMulticastList(priv->miniportAdapter, mclist, mc_count * ETH_ALEN));
			if (err) {
				printk("%s%s: Error %d setting multicast list.\n", (err == -EBUSY) ? KERN_DEBUG : KERN_ERR,
						dev->name, err);
				mc_count = 0;
			}

			OsFree(mclist);
		}
	}

	priv->mc_count = mc_count;

	return err;
	//printk("%s: out\n", __FUNCTION__, dev);
}

__shimcall__
static void
netndis_hang_handler(struct net_device *dev)
{
	struct netndis_private *priv = ((struct net_device *)dev)->priv;

	if (!priv->hw_unavailable) {
		if(!test_and_set_bit(NETNDIS_WORK_RESET, &priv->work_scheduled)) {
			OsTestPrintf("%s: scheduled reset_work\n", __FUNCTION__);
			if(OsThreadSchedule(priv->osthrd, &priv->reset_work, (void (*)(void *))netndis_reset, dev) <= 0) {
				clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
			}
		}
	}
}

__shimcall__
static void
netndis_resetcomplete_handler(struct net_device *dev, int err, BOOLEAN addressingreset)
{
	struct netndis_private *priv = ((struct net_device *)dev)->priv;

	priv->reset_err = err;
	priv->reset_addressing = addressingreset;
	up(&priv->reset_complete_sem);
}


/* This must be called from user context, without locks held - use
 * schedule_work() */
static void netndis_reset(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err;
	BOOLEAN AddressingReset;
	volatile OSIRQL oldIrql;
	unsigned long irqlstate;

	OsTestPrintf("%s\n", __FUNCTION__);

	err = netndis_lock(priv, FALSE);
	if (err) {
		/* When the hardware becomes available again, whatever
		 * detects that is responsible for re-initializing
		 * it. So no need for anything further*/
		OsModuleUseCountDec();
		clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
		return;
	}

	netif_stop_queue(dev);

#if 0
	/* Shut off interrupts.  Depending on what state the hardware
	 * is in, this might not work, but we'll try anyway */
	hhh_set_irqmask(hw, 0);
	hhh_write_regn(hw, EVACK, 0xffff);
#endif

	priv->hw_unavailable++;
	priv->last_linkstatus = NDISWRAP_LINKSTATUS_LAST; /* firmware will have to reassociate */

	netndis_unlock(priv);

	oldIrql = OsRaiseIrql(OSIRQL_DISPATCH, &irqlstate);
	err = NdisWrapStatusToErrno(
			NdisWrapResetAdapter(priv->miniportAdapter, &AddressingReset));
	OsRestoreIrql(oldIrql, &irqlstate);
	if (err == -EWOULDBLOCK) {
		// wait until netndis_resetcomplete_handler is called
		down(&priv->reset_complete_sem);

		err = priv->reset_err;
		AddressingReset = priv->reset_addressing;
	}

	if (err) {
		printk(KERN_ERR "%s: netndis_reset: Error %d performing reset\n", dev->name, err);

		NdisWrapHaltAdapter(priv->miniportAdapter);

		err = NdisWrapStatusToErrno(
				NdisWrapInitializeAdapter(priv->miniportAdapter,
					netndis_status_handler,
					netndis_rx_handler,
					netndis_xmitcomplete_handler,
					netndis_resetcomplete_handler,
					netndis_hang_handler));
		if(err) {
			printk(KERN_ERR "%s: failed to reinitialize adapter (err = %d)\n",
					dev->name, err);
			OsModuleUseCountDec();
			clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
			return;
		}

		err = netndis_lock(priv, TRUE);

		priv->hw_unavailable--;
		if(!err)
			netndis_unlock(priv);
		else {
			printk(KERN_ERR "%s: netndis_reset: Error %d locking (1)\n", dev->name, err);
		}
		OsModuleUseCountDec();
		clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
		return;
	}

	err = netndis_lock(priv, TRUE);
	if(err)
		printk(KERN_ERR "%s: netndis_reset: Error %d locking (2)\n", dev->name, err);

	priv->cfg = (1 << NETNDIS_CFG_LAST) - 1;

	priv->hw_unavailable--;

	/* priv->open or priv->hw_unavailable might have changed while
	 * we dropped the lock */
	if (priv->open && (!priv->hw_unavailable)) {
		int err = __netndis_up(dev);
		if (err) {
			printk(KERN_ERR "%s: netndis_reset: Error %d reenabling card\n",
			       dev->name, err);
		} else
			dev->trans_start = jiffies;
	}

	if(!err)
		netndis_unlock(priv);

	OsModuleUseCountDec();
	clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
	return;
}

/********************************************************************/
/* Wireless extensions support                                      */
/********************************************************************/

static int netndis_ioctl_getiwrange(struct net_device *dev, struct iw_point *rrq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	int mode;
	struct iw_range range;
	int numrates;
	int i, k;

	TRACE_ENTER(dev->name);

	err = verify_area(VERIFY_WRITE, rrq->pointer, sizeof(range));
	if (err)
		return err;

	rrq->length = sizeof(range);

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	mode = priv->iw_mode;
	netndis_unlock(priv);

	memset(&range, 0, sizeof(range));

	/* Much of this shamelessly taken from wvlan_cs.c. No idea
	 * what it all means -dgibson */
#if WIRELESS_EXT > 10
	range.we_version_compiled = WIRELESS_EXT;
	range.we_version_source = 11;
#endif /* WIRELESS_EXT > 10 */

	range.min_nwid = range.max_nwid = 0; /* We don't use nwids */

	/* Set available channels/frequencies */
	range.num_channels = NUM_CHANNELS;
	k = 0;
	for (i = 0; i < NUM_CHANNELS; i++) {
		if (priv->channel_mask & (1 << i)) {
			range.freq[k].i = i + 1;
			range.freq[k].m = channel_frequency[i] * 100000;
			range.freq[k].e = 1;
			k++;
		}
		
		if (k >= IW_MAX_FREQUENCIES)
			break;
	}
	range.num_frequency = k;

	range.sensitivity = 3;

#if 0
	if ((mode == IW_MODE_ADHOC) && (priv->spy_number == 0)){
		/* Quality stats meaningless in ad-hoc mode */
		range.max_qual.qual = 0;
		range.max_qual.level = 0;
		range.max_qual.noise = 0;
#if WIRELESS_EXT > 11
		range.avg_qual.qual = 0;
		range.avg_qual.level = 0;
		range.avg_qual.noise = 0;
#endif /* WIRELESS_EXT > 11 */

	} else {
		range.max_qual.qual = 0x8b - 0x2f;
		range.max_qual.level = 0x2f - 0x95 - 1;
		range.max_qual.noise = 0x2f - 0x95 - 1;
#if WIRELESS_EXT > 11
		/* Need to get better values */
		range.avg_qual.qual = 0x24;
		range.avg_qual.level = 0xC2;
		range.avg_qual.noise = 0x9E;
#endif /* WIRELESS_EXT > 11 */
	}
#else
	range.max_qual.qual = 1;
	range.max_qual.level = 0;
	range.max_qual.noise = 0;
	range.sensitivity = 65535;
#endif

	err = netndis_hw_get_bitratelist(priv, &numrates,
					 range.bitrate,
#if WIRELESS_EXT <= 15
					 IW_MAX_BITRATES - 1 /* due to stupid bug in iwlist.c */
#else
					 IW_MAX_BITRATES
#endif
					 );
	if (err)
		return err;
	range.num_bitrates = numrates;
	
	/* Set an indication of the max TCP throughput in bit/s that we can
	 * expect using this interface. May be use for QoS stuff...
	 * Jean II */
	if(numrates > 2)
		range.throughput = 5 * 1000 * 1000;	/* ~5 Mb/s */
	else
		range.throughput = 1.5 * 1000 * 1000;	/* ~1.5 Mb/s */

	range.min_rts = 0;
	range.max_rts = 2347;
	range.min_frag = 256;
	range.max_frag = 2346;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	if (priv->has_wep) {
		range.max_encoding_tokens = NETNDIS_MAX_KEYS;

		range.encoding_size[0] = SMALL_KEY_SIZE;
		range.num_encoding_sizes = 1;

		if (priv->has_big_wep) {
			range.encoding_size[1] = LARGE_KEY_SIZE;
			range.num_encoding_sizes = 2;
		}
	} else {
		range.num_encoding_sizes = 0;
		range.max_encoding_tokens = 0;
	}
	netndis_unlock(priv);
		
	range.min_pmp = 0;
	range.max_pmp = 65535000;
	range.min_pmt = 0;
	range.max_pmt = 65535 * 1000;	/* ??? */
	range.pmp_flags = IW_POWER_PERIOD;
	range.pmt_flags = IW_POWER_TIMEOUT;
	range.pm_capa = IW_POWER_PERIOD | IW_POWER_TIMEOUT | IW_POWER_UNICAST_R;

	range.num_txpower = 0;
	//range.txpower[0] = 15; /* 15dBm */
	range.txpower_capa = IW_TXPOW_MWATT;

#if 0 // WIRELESS_EXT > 10
	range.retry_capa = IW_RETRY_LIMIT | IW_RETRY_LIFETIME;
	range.retry_flags = IW_RETRY_LIMIT;
	range.r_time_flags = IW_RETRY_LIFETIME;
	range.min_retry = 0;
	range.max_retry = 65535;	/* ??? */
	range.min_r_time = 0;
	range.max_r_time = 65535 * 1000;	/* ??? */
#endif /* WIRELESS_EXT > 10 */

	if (copy_to_user(rrq->pointer, &range, sizeof(range)))
		return -EFAULT;

	TRACE_EXIT(dev->name);

	return 0;
}

static int netndis_ioctl_setiwencode(struct net_device *dev, struct iw_point *erq)
{
	struct netndis_private *priv = dev->priv;
	int index = (erq->flags & IW_ENCODE_INDEX) - 1;
	int setindex = priv->tx_key;
	u16 xlen = 0;
	int err = 0;
	char keybuf[NETNDIS_MAX_KEY_SIZE];
	
	if (erq->pointer) {
		/* We actually have a key to set */
		if ( (erq->length < SMALL_KEY_SIZE) || (erq->length > NETNDIS_MAX_KEY_SIZE) )
			return -EINVAL;
		
		if (copy_from_user(keybuf, erq->pointer, erq->length))
			return -EFAULT;
	}
	
	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	
	if (erq->pointer) {
		if (erq->length > NETNDIS_MAX_KEY_SIZE) {
			err = -E2BIG;
			goto out;
		}
		
		if ( (erq->length > LARGE_KEY_SIZE)
		     || ( !priv->has_big_wep && (erq->length > SMALL_KEY_SIZE))  ) {
			err = -EINVAL;
			goto out;
		}
		
		if ((index < 0) || (index >= NETNDIS_MAX_KEYS))
			index = priv->tx_key;
		
		if (erq->length > SMALL_KEY_SIZE) {
			xlen = LARGE_KEY_SIZE;
		} else if (erq->length > 0) {
			xlen = SMALL_KEY_SIZE;
		} else
			xlen = 0;
		
		/* Switch on WEP if off */
		if (((priv->iw_encode & IW_ENCODE_DISABLED)) && (xlen > 0)) {
			setindex = index;
			priv->iw_encode &= ~IW_ENCODE_DISABLED;
		}
	} else {
		/* Important note : if the user do "iwconfig eth0 enc off",
		 * we will arrive there with an index of -1. This is valid
		 * but need to be taken care off... Jean II */
		if ((index < 0) || (index >= NETNDIS_MAX_KEYS)) {
			if((index != -1) || (erq->flags == 0)) {
				err = -EINVAL;
				goto out;
			}
		} else {
			/* Set the index : Check that the key is valid */
			if(priv->keys[index].len == 0) {
				err = -EINVAL;
				goto out;
			}
			setindex = index;
		}
	}
	
	priv->iw_encode = erq->flags & IW_ENCODE_MODE;

	if (erq->pointer) {
		priv->keys[index].len = cpu_to_le16(xlen);
		memset(priv->keys[index].data, 0, sizeof(priv->keys[index].data));
		memcpy(priv->keys[index].data, keybuf, erq->length);
	}
	priv->tx_key = setindex;

 out:
	netndis_unlock(priv);

	return err;
}

static int netndis_ioctl_getiwencode(struct net_device *dev, struct iw_point *erq)
{
	struct netndis_private *priv = dev->priv;
	int index = (erq->flags & IW_ENCODE_INDEX) - 1;
	u16 xlen = 0;
	char keybuf[NETNDIS_MAX_KEY_SIZE];
	int err;
	
	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	if ((index < 0) || (index >= NETNDIS_MAX_KEYS))
		index = priv->tx_key;

	erq->flags = priv->iw_encode;
	erq->flags |= index + 1;

	xlen = le16_to_cpu(priv->keys[index].len);

	erq->length = xlen;

	if (erq->pointer) {
		memcpy(keybuf, priv->keys[index].data, NETNDIS_MAX_KEY_SIZE);
	}
	
	netndis_unlock(priv);

	if (erq->pointer) {
		if (copy_to_user(erq->pointer, keybuf, xlen))
			return -EFAULT;
	}

	return 0;
}

static int netndis_ioctl_setbssid(struct net_device *dev, struct sockaddr *awrq)
{
	struct netndis_private *priv = dev->priv;
	int err;
	static const unsigned char bcast[ETH_ALEN] = { 255, 255, 255, 255, 255, 255 };

	if(awrq->sa_family != ARPHRD_ETHER)
		return -EINVAL;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	if (!memcmp(bcast, awrq->sa_data, ETH_ALEN)) {
		err = NdisWrapStatusToErrno(
			NdisWrapDisassociate(priv->miniportAdapter));
	} else {
		err = NdisWrapStatusToErrno(
			NdisWrapSetBSSID(priv->miniportAdapter, awrq->sa_data, ETH_ALEN));
	}

	netndis_unlock(priv);

	return 0;
}

static int netndis_ioctl_setessid(struct net_device *dev, struct iw_point *erq)
{
	struct netndis_private *priv = dev->priv;
	char essidbuf[IW_ESSID_MAX_SIZE+1];
	int err;

	memset(&essidbuf, 0, sizeof(essidbuf));

	if (erq->flags) {
		if (erq->length > IW_ESSID_MAX_SIZE)
			return -E2BIG;
		
		if (copy_from_user(&essidbuf, erq->pointer, erq->length))
			return -EFAULT;

		essidbuf[erq->length] = '\0';
	}

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	memcpy(priv->desired_essid, essidbuf, sizeof(priv->desired_essid));

	netndis_unlock(priv);

	return 0;
}

static int netndis_ioctl_getessid(struct net_device *dev, struct iw_point *erq)
{
	struct netndis_private *priv = dev->priv;
	char essidbuf[IW_ESSID_MAX_SIZE+1];
	int active;
	int err;

	TRACE_ENTER(dev->name);

	if(priv->desired_essid[0]) {
		memcpy(essidbuf, priv->desired_essid, sizeof(essidbuf));
	} else {
		err = netndis_hw_get_essid(priv, &active, essidbuf);
		if (err)
			return err;
	}

	erq->flags = 1;
	erq->length = strlen(essidbuf) + 1;
	if (erq->pointer)
		if (copy_to_user(erq->pointer, essidbuf, erq->length))
			return -EFAULT;

	TRACE_EXIT(dev->name);
	
	return 0;
}

static int netndis_ioctl_setnick(struct net_device *dev, struct iw_point *nrq)
{
	struct netndis_private *priv = dev->priv;
	char nickbuf[IW_ESSID_MAX_SIZE+1];
	int err;

	if (nrq->length > IW_ESSID_MAX_SIZE)
		return -E2BIG;

	memset(nickbuf, 0, sizeof(nickbuf));

	if (copy_from_user(nickbuf, nrq->pointer, nrq->length))
		return -EFAULT;

	nickbuf[nrq->length] = '\0';
	
	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	memcpy(priv->nick, nickbuf, sizeof(priv->nick));

	netndis_unlock(priv);

	return 0;
}

static int netndis_ioctl_getnick(struct net_device *dev, struct iw_point *nrq)
{
	struct netndis_private *priv = dev->priv;
	char nickbuf[IW_ESSID_MAX_SIZE+1];
	int err;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	memcpy(nickbuf, priv->nick, IW_ESSID_MAX_SIZE+1);
	netndis_unlock(priv);

	nrq->length = strlen(nickbuf)+1;

	if (copy_to_user(nrq->pointer, nickbuf, sizeof(nickbuf)))
		return -EFAULT;

	return 0;
}

static int netndis_ioctl_setfreq(struct net_device *dev, struct iw_freq *frq)
{
	struct netndis_private *priv = dev->priv;
	int chan = -1;
	int err;

	/* We can only use this in Ad-Hoc demo mode to set the operating
	 * frequency, or in IBSS mode to set the frequency where the IBSS
	 * will be created - Jean II */
	if (priv->iw_mode != IW_MODE_ADHOC)
		return -EOPNOTSUPP;

	if ( (frq->e == 0) && (frq->m <= 1000) ) {
		/* Setting by channel number */
		chan = frq->m;
	} else {
		/* Setting by frequency - search the table */
		int mult = 1;
		int i;

		for (i = 0; i < (6 - frq->e); i++)
			mult *= 10;

		for (i = 0; i < NUM_CHANNELS; i++)
			if (frq->m == (channel_frequency[i] * mult))
				chan = i+1;
	}

	if ( (chan < 1) || (chan > NUM_CHANNELS) ||
	     !(priv->channel_mask & (1 << (chan-1)) ) )
		return -EINVAL;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	priv->channel = chan;
	netndis_unlock(priv);

	return 0;
}

#if 0
static int netndis_ioctl_getsens(struct net_device *dev, struct iw_param *srq)
{
	struct netndis_private *priv = dev->priv;
	u16 val;
	int err;

	if (!priv->has_sensitivity)
		return -EOPNOTSUPP;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
#if 0
	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_CNFSYSTEMSCALE, &val);
#else
	/* XXX */
#endif
	netndis_unlock(priv);

	if (err)
		return err;

	srq->value = val;
	srq->fixed = 0; /* auto */

	return 0;
}

static int netndis_ioctl_setsens(struct net_device *dev, struct iw_param *srq)
{
	struct netndis_private *priv = dev->priv;
	int val = srq->value;
	int err;

	if (!priv->has_sensitivity)
		return -EOPNOTSUPP;

	if ((val < 1) || (val > 3))
		return -EINVAL;
	
	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	priv->ap_density = val;
	netndis_unlock(priv);

	return 0;
}
#endif

static int netndis_ioctl_setrts(struct net_device *dev, struct iw_param *rrq)
{
	struct netndis_private *priv = dev->priv;
	int val = rrq->value;
	int err;

	if (rrq->disabled)
		val = 2347;

	if ( (val < 0) || (val > 2347) )
		return -EINVAL;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	priv->rts_thresh = val;
	netndis_unlock(priv);

	return 0;
}

static int netndis_ioctl_getrts(struct net_device *dev, struct iw_param *frq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	ULONG val;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	
	err = NdisWrapStatusToErrno(
			NdisWrapGetRtsThreshHold(priv->miniportAdapter, &val));
	if (err)
		val = 2347;

	frq->value = val;
	frq->disabled = (val >= 2347);
	frq->fixed = 1;

	netndis_unlock(priv);
	
	return err;
}

static int netndis_ioctl_setfrag(struct net_device *dev, struct iw_param *frq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	if (frq->disabled)
		priv->frag_thresh = 2346;
	else {
		if ( (frq->value < 256) || (frq->value > 2346) )
			err = -EINVAL;
		else
			priv->frag_thresh = frq->value & ~0x1; /* must be even */
	}

	netndis_unlock(priv);

	return err;
}

static int netndis_ioctl_getfrag(struct net_device *dev, struct iw_param *frq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	ULONG val;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	
	err = NdisWrapStatusToErrno(
			NdisWrapGetFragThreshHold(priv->miniportAdapter, &val));
	if (err)
		val = 2346;

	frq->value = val;
	frq->disabled = (val >= 2346);
	frq->fixed = 1;

	netndis_unlock(priv);
	
	return err;
}

#if 0
static int netndis_ioctl_settxpower(struct net_device *dev, struct iw_param *trq)
{
	struct netndis_private *priv = dev->priv;
	int err;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	err = __netndis_hw_set_txpower(priv, trq);
	if(!err)
		priv->txpowermode = *trq;

	netndis_unlock(priv);

	return err;
}
#endif

static int netndis_ioctl_gettxpower(struct net_device *dev, struct iw_param *trq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	ULONG UVal;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	err = NdisWrapStatusToErrno(
			NdisWrapGetTxPowerLevel(priv->miniportAdapter, &UVal));
	if (err)
		goto out;

	trq->value = UVal;
	trq->fixed = 1;
	trq->disabled = 0;
	trq->flags = IW_TXPOW_MWATT;

 out:
	netndis_unlock(priv);

	return err;
}

static int netndis_ioctl_setrate(struct net_device *dev, struct iw_param *rrq)
{
	struct netndis_private *priv = dev->priv;
	int err;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	err = __netndis_hw_set_bitrate(priv, rrq);
	if(!err)
		priv->bitratemode = *rrq;

	netndis_unlock(priv);

	return err;
}

static int netndis_ioctl_getrate(struct net_device *dev, struct iw_param *rrq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	ULONG UVal;
#ifdef DLDR_DESIRED_RATES_OID_SUPPORTED
	unsigned char desiredrates[16];
#endif

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	err = NdisWrapStatusToErrno(
			NdisWrapGetLinkSpeed(priv->miniportAdapter, &UVal));
	if (err)
		goto out;

	rrq->value = UVal * 100;

#ifdef DLDR_DESIRED_RATES_OID_SUPPORTED
	rrq->fixed = NdisWrapStatusToErrno(
			NdisWrapGetDesiredRates(priv->miniportAdapter, desiredrates, sizeof(desiredrates))) || (desiredrates[0] && !desiredrates[1]);
#else
	rrq->fixed = 1;
#endif
	rrq->disabled = 0;

 out:
	netndis_unlock(priv);

	return err;
}

static int netndis_ioctl_setpower(struct net_device *dev, struct iw_param *prq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	if (prq->disabled) {
		priv->pm_on = 0;
	} else {
		switch (prq->flags & IW_POWER_MODE) {
		case IW_POWER_UNICAST_R:
			priv->pm_mcast = 0;
			priv->pm_on = 1;
			break;
		case IW_POWER_ALL_R:
			priv->pm_mcast = 1;
			priv->pm_on = 1;
			break;
		case IW_POWER_ON:
			/* No flags : but we may have a value - Jean II */
			break;
		default:
			err = -EINVAL;
		}
		if (err)
			goto out;
		
		if (prq->flags & IW_POWER_TIMEOUT) {
			priv->pm_on = 1;
			priv->pm_timeout = prq->value / 1000;
		}
		if (prq->flags & IW_POWER_PERIOD) {
			priv->pm_on = 1;
			priv->pm_period = prq->value / 1000;
		}
		/* It's valid to not have a value if we are just toggling
		 * the flags... Jean II */
		if(!priv->pm_on) {
			err = -EINVAL;
			goto out;
		}			
	}

 out:
	netndis_unlock(priv);

	return err;
}

static int netndis_ioctl_getpower(struct net_device *dev, struct iw_param *prq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	
#if 0
	err = hhh_read_wordrec(hw, USER_BAP,
				  HHH_RID_CNFMAXSLEEPDURATION, &period);
	if (err)
		goto out;

	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_CNFPMHOLDOVERDURATION, &timeout);
	if (err)
		goto out;
#endif

	prq->disabled = !priv->pm_on;
	/* Note : by default, display the period */
	if ((prq->flags & IW_POWER_TYPE) == IW_POWER_TIMEOUT) {
		prq->flags = IW_POWER_TIMEOUT;
		prq->value = 0; // XXX timeout * 1000;
	} else {
		prq->flags = IW_POWER_PERIOD;
		prq->value = 0; // XXX period * 1000;
	}
	if (priv->pm_mcast)
		prq->flags |= IW_POWER_ALL_R;
	else
		prq->flags |= IW_POWER_UNICAST_R;

	netndis_unlock(priv);

	return err;
}

#if 0 // WIRELESS_EXT > 10
static int netndis_ioctl_getretry(struct net_device *dev, struct iw_param *rrq)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	u16 short_limit, long_limit, lifetime;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;
	
#if 0
	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_SHORTRETRYLIMIT,
				  &short_limit);
	if (err)
		goto out;

	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_LONGRETRYLIMIT,
				  &long_limit);
	if (err)
		goto out;

	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_MAXTRANSMITLIFETIME,
				  &lifetime);
	if (err)
		goto out;
#else
	short_limit = long_limit = lifetime = 0; /* XXX */
#endif

	rrq->disabled = 0;		/* Can't be disabled */

	/* Note : by default, display the retry number */
	if ((rrq->flags & IW_RETRY_TYPE) == IW_RETRY_LIFETIME) {
		rrq->flags = IW_RETRY_LIFETIME;
		rrq->value = lifetime * 1000;	/* ??? */
	} else {
		/* By default, display the min number */
		if ((rrq->flags & IW_RETRY_MAX)) {
			rrq->flags = IW_RETRY_LIMIT | IW_RETRY_MAX;
			rrq->value = long_limit;
		} else {
			rrq->flags = IW_RETRY_LIMIT;
			rrq->value = short_limit;
			if(short_limit != long_limit)
				rrq->flags |= IW_RETRY_MIN;
		}
	}

 out:
	netndis_unlock(priv);

	return err;
}
#endif /* WIRELESS_EXT > 10 */

/* Spy is used for link quality/strength measurements in Ad-Hoc mode
 * Jean II */
static int netndis_ioctl_setspy(struct net_device *dev, struct iw_point *srq)
{
	struct netndis_private *priv = dev->priv;
	struct sockaddr address[IW_MAX_SPY];
	int number = srq->length;
	int i;
	int err = 0;

	/* Check the number of addresses */
	if (number > IW_MAX_SPY)
		return -E2BIG;

	/* Get the data in the driver */
	if (srq->pointer) {
		if (copy_from_user(address, srq->pointer,
				   sizeof(struct sockaddr) * number))
			return -EFAULT;
	}

	/* Make sure nobody mess with the structure while we do */
	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	/* netndis_lock() doesn't disable interrupts, so make sure the
	 * interrupt rx path don't get confused while we copy */
	priv->spy_number = 0;

	if (number > 0) {
		/* Extract the addresses */
		for (i = 0; i < number; i++)
			memcpy(priv->spy_address[i], address[i].sa_data,
			       ETH_ALEN);
		/* Reset stats */
		memset(priv->spy_stat, 0,
		       sizeof(struct iw_quality) * IW_MAX_SPY);
		/* Set number of addresses */
		priv->spy_number = number;
	}

	/* Now, let the others play */
	netndis_unlock(priv);

	return err;
}

static int netndis_ioctl_getspy(struct net_device *dev, struct iw_point *srq)
{
	struct netndis_private *priv = dev->priv;
	struct sockaddr address[IW_MAX_SPY];
	struct iw_quality spy_stat[IW_MAX_SPY];
	int number;
	int i;
	int err;

	err = netndis_lock(priv, FALSE);
	if (err)
		return err;

	number = priv->spy_number;
	if ((number > 0) && (srq->pointer)) {
		/* Create address struct */
		for (i = 0; i < number; i++) {
			memcpy(address[i].sa_data, priv->spy_address[i],
			       ETH_ALEN);
			address[i].sa_family = AF_UNIX;
		}
		/* Copy stats */
		/* In theory, we should disable irqs while copying the stats
		 * because the rx path migh update it in the middle...
		 * Bah, who care ? - Jean II */
		memcpy(&spy_stat, priv->spy_stat,
		       sizeof(struct iw_quality) * IW_MAX_SPY);
		for (i=0; i < number; i++)
			priv->spy_stat[i].updated = 0;
	}

	netndis_unlock(priv);

	/* Push stuff to user space */
	srq->length = number;
	if(copy_to_user(srq->pointer, address,
			 sizeof(struct sockaddr) * number))
		return -EFAULT;
	if(copy_to_user(srq->pointer + (sizeof(struct sockaddr)*number),
			&spy_stat, sizeof(struct iw_quality) * number))
		return -EFAULT;

	return 0;
}

static int netndis_ioctl_get_aplist(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_point *dwrq,
			   char *extra)
{
	struct netndis_private *priv = dev->priv;
	struct sockaddr *address = (struct sockaddr *) extra;
	struct iw_quality qual[IW_MAX_AP];
	int i, err;
	PVOID BSSList;
	ULONG BSSListNumEntries;
	LONG Rssi;

	if(capable(CAP_NET_ADMIN)) {
		priv->scan_timestamp = jiffies;
		NdisWrapSetBSSIDListScan(priv->miniportAdapter);
	}

	err = NdisWrapStatusToErrno(
			NdisWrapGetBSSIDList(priv->miniportAdapter, &BSSList, &BSSListNumEntries));
	if(err)
		return err;

	for (i = 0; i < BSSListNumEntries; i++) {

		NdisWrapGetBSSIDListEntryMacAddress(BSSList, i, address[i].sa_data);
		address[i].sa_family = ARPHRD_ETHER;


		NdisWrapGetBSSIDListEntryRssi(BSSList, i, &Rssi);

		qual[i].level = 0x100 + Rssi;
		qual[i].qual = 1;
		qual[i].noise = 0x100 - 200;
		qual[i].updated = 2;
	}

#if 0
	if (!i) {
		StatusRid status_rid;		/* Card status info */
		readStatusRid(local, &status_rid);
		for (i = 0;
		     i < min(IW_MAX_AP, 4) &&
			     (status_rid.bssid[i][0]
			      & status_rid.bssid[i][1]
			      & status_rid.bssid[i][2]
			      & status_rid.bssid[i][3]
			      & status_rid.bssid[i][4]
			      & status_rid.bssid[i][5])!=-1 &&
			     (status_rid.bssid[i][0]
			      | status_rid.bssid[i][1]
			      | status_rid.bssid[i][2]
			      | status_rid.bssid[i][3]
			      | status_rid.bssid[i][4]
			      | status_rid.bssid[i][5]);
		     i++) {
			memcpy(address[i].sa_data,
			       status_rid.bssid[i], ETH_ALEN);
			address[i].sa_family = ARPHRD_ETHER;
		}
	} else
#endif
	{
		dwrq->flags = 1; /* Should be define'd */
		memcpy(extra + sizeof(struct sockaddr)*i,
		       &qual,  sizeof(struct iw_quality)*i);
	}
	dwrq->length = i;

	NdisWrapFreeBSSIDList(BSSList);

	return 0;
}

#ifdef SIOCGIWSCAN
static int netndis_ioctl_get_scan(struct net_device *dev,
			 struct iw_request_info *info,
			 struct iw_point *dwrq,
			 char *extra)
{
	struct netndis_private *priv = dev->priv;
	PVOID BSSList;
	ULONG BSSListNumEntries;
	int i, j;
	ULONG UVal;
	LONG Rssi;
	char ssid[IW_ESSID_MAX_SIZE+1];
	UCHAR SupportedRates[16];
	int err;
	char *current_ev = extra;
	char *end_buf = extra + IW_SCAN_MAX_DATA;
	struct iw_event iwe; /* Temporary buffer */
	char *current_val; /* For rates */

	/* When we are associated again, the scan has surely finished.
	 * Just in case, let's make sure enough time has elapsed since
	 * we started the scan. - Javier */
	if(priv->scan_timestamp && time_before(jiffies,priv->scan_timestamp+3*HZ)) {
		/* Important note : we don't want to block the caller
		 * until results are ready for various reasons.
		 * First, managing wait queues is complex and racy
		 * (there may be multiple simultaneous callers).
		 * Second, we grab some rtnetlink lock before comming
		 * here (in dev_ioctl()).
		 * Third, the caller can wait on the Wireless Event
		 * - Jean II */
		return -EAGAIN;
	}
	priv->scan_timestamp = 0;

	/* There's only a race with proc_BSSList_open(), but its
	 * consequences are begnign. So I don't bother fixing it - Javier */

	err = NdisWrapStatusToErrno(
			NdisWrapGetBSSIDList(priv->miniportAdapter, &BSSList, &BSSListNumEntries));
	if(err || !BSSListNumEntries) {
		/* Client error, no scan results...
		 * The caller need to restart the scan. */
		NdisWrapFreeBSSIDList(BSSList);
		return -ENODATA;
	}

	for (i = 0; i < BSSListNumEntries; i++) {

		/* First entry *MUST* be the AP MAC address */
		NdisWrapGetBSSIDListEntryMacAddress(BSSList, i, iwe.u.ap_addr.sa_data);
		iwe.cmd = SIOCGIWAP;
		iwe.u.ap_addr.sa_family = ARPHRD_ETHER;

		current_ev = iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_ADDR_LEN);

		/* Other entries will be displayed in the order we give them */

		/* Add the ESSID */
		NdisWrapGetBSSIDListEntrySSID(BSSList, i, ssid, sizeof(ssid));
		iwe.u.data.length = strlen(ssid);
		if(iwe.u.data.length > 32)
			iwe.u.data.length = 32;
		iwe.cmd = SIOCGIWESSID;
		iwe.u.data.flags = 1;
		current_ev = iwe_stream_add_point(current_ev, end_buf, &iwe, ssid);

		/* Add mode */
		NdisWrapGetBSSIDListEntryInfrastructureMode(BSSList, i, (ndiswrapiwmode_t*)&j);
		iwe.cmd = SIOCGIWMODE;
		iwe.u.mode = j;
		current_ev = iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_UINT_LEN);

		/* Add frequency */
		NdisWrapGetBSSIDListEntryConfiguration(BSSList, i, NULL, NULL, &UVal, NULL, NULL, NULL);
		iwe.cmd = SIOCGIWFREQ;
		iwe.u.freq.m = UVal * 100;
		iwe.u.freq.e = 1;
		current_ev = iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_FREQ_LEN);

		/* Add quality statistics */
		NdisWrapGetBSSIDListEntryRssi(BSSList, i, &Rssi);
		iwe.cmd = IWEVQUAL;
		iwe.u.qual.level = 0x100 + Rssi;
		iwe.u.qual.noise = 0x100 - 200;
		iwe.u.qual.qual = 1;
		current_ev = iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_QUAL_LEN);

		/* Add encryption capability */
		NdisWrapGetBSSIDListEntryPrivacy(BSSList, i, &UVal);
		iwe.cmd = SIOCGIWENCODE;
		if(UVal)
			iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
		else
			iwe.u.data.flags = IW_ENCODE_DISABLED;
		iwe.u.data.length = 0;
		current_ev = iwe_stream_add_point(current_ev, end_buf, &iwe, ssid);

		/* Rate : stuffing multiple values in a single event require a bit
		 * more of magic - Jean II */
		current_val = current_ev + IW_EV_LCP_LEN;

		iwe.cmd = SIOCGIWRATE;
		/* Those two flags are ignored... */
		iwe.u.bitrate.fixed = iwe.u.bitrate.disabled = 0;
		NdisWrapGetBSSIDListEntrySupportedRates(BSSList, i, SupportedRates, sizeof(SupportedRates));
		for(j = 0 ; j < sizeof(SupportedRates) ; j++) {
			/* NULL terminated */
			if(SupportedRates[j] == 0)
				break;
			/* Bit rate given in 500 kb/s units (+ 0x80) */
			iwe.u.bitrate.value = ((SupportedRates[j] & 0x7f) * 500000);
			/* Add new value to event */
			current_val = iwe_stream_add_value(current_ev, current_val, end_buf, &iwe, IW_EV_PARAM_LEN);
		}
		/* Check if we added any event */
		if((current_val - current_ev) > IW_EV_LCP_LEN)
			current_ev = current_val;
	}
	/* Length of data */
	dwrq->length = (current_ev - extra);
	dwrq->flags = 0;	/* todo */

	NdisWrapFreeBSSIDList(BSSList);

	return 0;
}
#endif /* SIOCGIWSCAN */

static int netndis_ioctl_ethtool(struct net_device *dev, void *useraddr)
{
	u32 ethcmd;
		
	if (copy_from_user(&ethcmd, useraddr, sizeof(ethcmd)))
		return -EFAULT;
	
	switch (ethcmd) {
#ifdef ETHTOOL_GDRVINFO
	case ETHTOOL_GDRVINFO: {
		struct ethtool_drvinfo info = {ETHTOOL_GDRVINFO};
		strncpy(info.driver, THIS_MODULE->name, sizeof(info.driver)-1);
		strncpy(info.version, DLDRNAME" "DLDRVERSION" - "DLDRDESC, sizeof(info.version)-1);
		if (copy_to_user(useraddr, &info, sizeof(info)))
			return -EFAULT;
		return 0;
						   }
#endif
	
	}
	
	return -EOPNOTSUPP;
}

static int
netndis_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	struct netndis_private *priv = dev->priv;
	struct iwreq *wrq = (struct iwreq *)rq;
	int err = 0;
	int tmp;

	TRACE_ENTER(dev->name);

	/* In theory, we could allow most of the the SET stuff to be
	 * done. In practice, the lapse of time at startup when the
	 * card is not ready is very short, so why bother...  Note
	 * that netif_device_present is different from up/down
	 * (ifconfig), when the device is not yet up, it is usually
	 * already ready...  Jean II */
	if (!netif_device_present(dev))
		return -ENODEV;

	switch (cmd) {
	case SIOCGIWNAME:
		OsDebugPrintf("%s: SIOCGIWNAME\n", __FUNCTION__);
		{
			ndiswrapnetworktype_t   NetworkTypeInUse;
			char *networktype = "IEEE 802.11-DS";

			if(!NdisWrapStatusToErrno(
				NdisWrapGetNetworkType(priv->miniportAdapter, &NetworkTypeInUse))) {
				if(NetworkTypeInUse == NDISWRAP_NETWORKTYPE_FH) {
					networktype = "IEEE 802.11-FH";
				} else if(NetworkTypeInUse == NDISWRAP_NETWORKTYPE_DS) {
					//networktype = "IEEE 802.11-DS";
				} else if(NetworkTypeInUse == NDISWRAP_NETWORKTYPE_OFDM5) {
					networktype = "IEEE 802.11a";
				} else if(NetworkTypeInUse == NDISWRAP_NETWORKTYPE_OFDM24) {
					networktype = "IEEE 802.11g";
				}
			}
			strncpy(wrq->u.name, networktype, sizeof(wrq->u.name));
		}
		break;
		
	case SIOCGIWAP:
		OsDebugPrintf("%s: SIOCGIWAP\n", __FUNCTION__);
		wrq->u.ap_addr.sa_family = ARPHRD_ETHER;
		err = netif_carrier_ok(dev) ? netndis_hw_get_bssid(priv, wrq->u.ap_addr.sa_data) : -ENETDOWN;
		break;

	case SIOCSIWAP:
		OsDebugPrintf("%s: SIOCSIWAP\n", __FUNCTION__);
		err = netndis_ioctl_setbssid(dev, &wrq->u.ap_addr);
		break;

	case SIOCGIWRANGE:
		OsDebugPrintf("%s: SIOCGIWRANGE\n", __FUNCTION__);
		err = netndis_ioctl_getiwrange(dev, &wrq->u.data);
		break;

	case SIOCSIWMODE:
		OsDebugPrintf("%s: SIOCSIWMODE\n", __FUNCTION__);
		err = netndis_lock(priv, FALSE);
		if (err)
			return err;
		switch (wrq->u.mode) {
		case IW_MODE_ADHOC:
		case IW_MODE_INFRA:
			priv->iw_mode = wrq->u.mode;
			set_bit(NETNDIS_CFG_MODE, &priv->cfg);
			break;

		default:
			err = -EINVAL;
			break;
		}
		netndis_unlock(priv);
		break;

	case SIOCGIWMODE:
		OsDebugPrintf("%s: SIOCGIWMODE\n", __FUNCTION__);
		err = netndis_lock(priv, FALSE);
		if (err)
			return err;
		wrq->u.mode = priv->iw_mode;
		netndis_unlock(priv);
		break;

	case SIOCSIWENCODE:
		OsDebugPrintf("%s: SIOCSIWENCODE\n", __FUNCTION__);
		if (!priv->has_wep) {
			err = -EOPNOTSUPP;
			break;
		}

		err = netndis_ioctl_setiwencode(dev, &wrq->u.encoding);
		if (!err) {
			set_bit(NETNDIS_CFG_ENC, &priv->cfg);
			set_bit(NETNDIS_CFG_KEYS, &priv->cfg);
			set_bit(NETNDIS_CFG_AUTHMODE, &priv->cfg);
		}
		break;

	case SIOCGIWENCODE:
		OsDebugPrintf("%s: SIOCGIWENCODE\n", __FUNCTION__);
		if (!priv->has_wep) {
			err = -EOPNOTSUPP;
			break;
		}

		if (!capable(CAP_NET_ADMIN)) {
			err = -EPERM;
			break;
		}

		err = netndis_ioctl_getiwencode(dev, &wrq->u.encoding);
		break;

	case SIOCSIWESSID:
		OsDebugPrintf("%s: SIOCSIWESSID\n", __FUNCTION__);
		err = netndis_ioctl_setessid(dev, &wrq->u.essid);
		if (!err) {
			set_bit(NETNDIS_CFG_ESSID, &priv->cfg);
		}
		break;

	case SIOCGIWESSID:
		OsDebugPrintf("%s: SIOCGIWESSID\n", __FUNCTION__);
		err = netndis_ioctl_getessid(dev, &wrq->u.essid);
		break;

	case SIOCSIWNICKN:
		OsDebugPrintf("%s: SIOCSIWNICKN\n", __FUNCTION__);
		err = netndis_ioctl_setnick(dev, &wrq->u.data);
		if (!err) {
			set_bit(NETNDIS_CFG_NICK, &priv->cfg);
		}
		break;

	case SIOCGIWNICKN:
		OsDebugPrintf("%s: SIOCGIWNICKN\n", __FUNCTION__);
		err = netndis_ioctl_getnick(dev, &wrq->u.data);
		break;

	case SIOCGIWFREQ:
		OsDebugPrintf("%s: SIOCGIWFREQ\n", __FUNCTION__);
		tmp = netndis_hw_get_freq(priv);
		if (tmp < 0) {
			err = tmp;
		} else {
			wrq->u.freq.m = tmp;
			wrq->u.freq.e = 1;
		}
		break;

	case SIOCSIWFREQ:
		OsDebugPrintf("%s: SIOCSIWFREQ\n", __FUNCTION__);
		err = netndis_ioctl_setfreq(dev, &wrq->u.freq);
		if (!err) {
			set_bit(NETNDIS_CFG_FREQ, &priv->cfg);
		}
		break;

	case SIOCGIWSENS:
		OsDebugPrintf("%s: SIOCGIWSENS\n", __FUNCTION__);
		err = -EOPNOTSUPP;
		//err = netndis_ioctl_getsens(dev, &wrq->u.sens);
		break;

	case SIOCSIWSENS:
		OsDebugPrintf("%s: SIOCSIWSENS\n", __FUNCTION__);
		err = -EOPNOTSUPP;
#if 0
		err = netndis_ioctl_setsens(dev, &wrq->u.sens);
		if (!err) {
			set_bit(NETNDIS_CFG_SENS, &priv->cfg);
		}
#endif
		break;

	case SIOCGIWRTS:
		OsDebugPrintf("%s: SIOCGIWRTS\n", __FUNCTION__);
#if 1
		err = netndis_ioctl_getrts(dev, &wrq->u.rts);
#else
		wrq->u.rts.value = priv->rts_thresh;
		wrq->u.rts.disabled = (wrq->u.rts.value == 2347);
		wrq->u.rts.fixed = 1;
#endif
		break;

	case SIOCSIWRTS:
		OsDebugPrintf("%s: SIOCSIWRTS\n", __FUNCTION__);
		err = netndis_ioctl_setrts(dev, &wrq->u.rts);
		if (!err) {
			set_bit(NETNDIS_CFG_RTS, &priv->cfg);
		}
		break;

	case SIOCSIWFRAG:
		OsDebugPrintf("%s: SIOCSIWFRAG\n", __FUNCTION__);
		err = netndis_ioctl_setfrag(dev, &wrq->u.frag);
		if (!err) {
			set_bit(NETNDIS_CFG_FRAG, &priv->cfg);
		}
		break;

	case SIOCGIWFRAG:
		OsDebugPrintf("%s: SIOCGIWFRAG\n", __FUNCTION__);
		err = netndis_ioctl_getfrag(dev, &wrq->u.frag);
		break;

	case SIOCSIWRATE:
		OsDebugPrintf("%s: SIOCSIWRATE\n", __FUNCTION__);
		err = netndis_ioctl_setrate(dev, &wrq->u.bitrate);
		if (!err) {
			set_bit(NETNDIS_CFG_RATE, &priv->cfg);
		}
		break;

	case SIOCGIWRATE:
		OsDebugPrintf("%s: SIOCGIWRATE\n", __FUNCTION__);
		err = netndis_ioctl_getrate(dev, &wrq->u.bitrate);
		break;

	case SIOCSIWPOWER:
		OsDebugPrintf("%s: SIOCSIWPOWER\n", __FUNCTION__);
		err = netndis_ioctl_setpower(dev, &wrq->u.power);
		if (!err) {
			set_bit(NETNDIS_CFG_PM, &priv->cfg);
		}
		break;

	case SIOCGIWPOWER:
		OsDebugPrintf("%s: SIOCGIWPOWER\n", __FUNCTION__);
		err = netndis_ioctl_getpower(dev, &wrq->u.power);
		break;

	case SIOCGIWTXPOW:
		OsDebugPrintf("%s: SIOCGIWTXPOW\n", __FUNCTION__);
		err = netndis_ioctl_gettxpower(dev, &wrq->u.txpower);
		break;

#if WIRELESS_EXT > 10
	case SIOCSIWRETRY:
		OsDebugPrintf("%s: SIOCSIWRETRY\n", __FUNCTION__);
		err = -EOPNOTSUPP;
		break;

	case SIOCGIWRETRY:
		OsDebugPrintf("%s: SIOCGIWRETRY\n", __FUNCTION__);
		//err = netndis_ioctl_getretry(dev, &wrq->u.retry);
		err = -EOPNOTSUPP;
		break;
#endif /* WIRELESS_EXT > 10 */

	case SIOCSIWSPY:
		OsDebugPrintf("%s: SIOCSIWSPY\n", __FUNCTION__);
		err = netndis_ioctl_setspy(dev, &wrq->u.data);
		break;

	case SIOCGIWSPY:
		OsDebugPrintf("%s: SIOCGIWSPY\n", __FUNCTION__);
		err = netndis_ioctl_getspy(dev, &wrq->u.data);
		break;

	case SIOCGIWAPLIST:
		OsDebugPrintf("%s: SIOCGIWAPLIST\n", __FUNCTION__);
		{
			char buffer[IW_MAX_AP * (sizeof(struct sockaddr) +
						  sizeof(struct iw_quality))];
			if (wrq->u.data.pointer) {
				err = netndis_ioctl_get_aplist(dev, NULL,
						     &(wrq->u.data), buffer);
				if (copy_to_user(wrq->u.data.pointer,
						 buffer,
						 (wrq->u.data.length *
						  (sizeof(struct sockaddr) +
						   sizeof(struct iw_quality)))
						 ))
					err = -EFAULT;
			}
		}
		break;

#ifdef SIOCSIWSCAN
	case SIOCSIWSCAN:
		OsDebugPrintf("%s: SIOCSIWSCAN\n", __FUNCTION__);
		priv->scan_timestamp = jiffies;
		err = NdisWrapStatusToErrno(
				NdisWrapSetBSSIDListScan(priv->miniportAdapter));
		break;
#endif

#ifdef SIOCGIWSCAN
	case SIOCGIWSCAN:
		OsDebugPrintf("%s: SIOCGIWSCAN\n", __FUNCTION__);
		{
			char *buffer;

			buffer = kmalloc(IW_SCAN_MAX_DATA, GFP_KERNEL);
			if(!buffer)
				err = -ENOMEM;
			else {

				if (wrq->u.data.pointer) {
					err = netndis_ioctl_get_scan(dev, NULL,
							&(wrq->u.data), buffer);
					if (copy_to_user(wrq->u.data.pointer,
								buffer,
								wrq->u.data.length))
						err = -EFAULT;
				}
				kfree(buffer);
			}
		}
		break;
#endif

	case SIOCGIWPRIV:
		OsDebugPrintf("%s: SIOCGIWPRIV\n", __FUNCTION__);
		if (wrq->u.data.pointer) {
			struct iw_priv_args privtab[] = {
				{ SIOCIWFIRSTPRIV + 0x0, 0, 0, "force_reset" },
				{ SIOCIWFIRSTPRIV + 0x1, 0, 0, "card_reset" },
			};

			err = verify_area(VERIFY_WRITE, wrq->u.data.pointer, sizeof(privtab));
			if (err)
				break;
			
			wrq->u.data.length = sizeof(privtab) / sizeof(privtab[0]);
			if (copy_to_user(wrq->u.data.pointer, privtab, sizeof(privtab)))
				err = -EFAULT;
		}
		break;
	       
	case SIOCIWFIRSTPRIV + 0x0: /* force_reset */
		OsDebugPrintf("%s: SIOCIWFIRSTPRIV+0x0 force_reset\n", __FUNCTION__);
	case SIOCIWFIRSTPRIV + 0x1: /* card_reset */
		OsDebugPrintf("%s: SIOCIWFIRSTPRIV+0x1 card_reset\n", __FUNCTION__);
		if (!capable(CAP_NET_ADMIN)) {
			err = -EPERM;
			break;
		}
		
		printk(KERN_DEBUG "%s: Force scheduling reset!\n", dev->name);

		if(!test_and_set_bit(NETNDIS_WORK_RESET, &priv->work_scheduled)) {
			OsTestPrintf("%s: scheduled reset_work\n", __FUNCTION__);
			if(OsThreadSchedule(priv->osthrd, &priv->reset_work, (void (*)(void *))netndis_reset, dev) <= 0) {
				clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
			}
		}
		break;

#ifdef SIOCETHTOOL
	case SIOCETHTOOL:
		OsDebugPrintf("%s: SIOCETHTOOL\n", __FUNCTION__);
		err = netndis_ioctl_ethtool(dev, (void *) rq->ifr_data);
		break;
#endif

	default:
		//printk(KERN_ERR"%s: unsupported ioctl %x\n", __FUNCTION__, cmd);
		err = -EOPNOTSUPP;
	}
	
	if (!err && priv->cfg) {

		err = netndis_lock(priv, FALSE);
		if (!err) {

			err = __netndis_configure(dev, netif_running(dev));
			if (err) {
#if 1
				printk(KERN_WARNING "%s: Unable to configure card; resetting...\n", dev->name);
				if(!test_and_set_bit(NETNDIS_WORK_RESET, &priv->work_scheduled)) {
					OsTestPrintf("%s: scheduled reset_work\n", __FUNCTION__);
					if(OsThreadSchedule(priv->osthrd, &priv->reset_work, (void (*)(void *))netndis_reset, dev) <= 0) {
						clear_bit(NETNDIS_WORK_RESET, &priv->work_scheduled);
					}
				}
#endif
				err = 0;
			}

			netndis_unlock(priv);
		}
	}		

	TRACE_EXIT(dev->name);

	return err;
}

static int netndis_get_hwinst(char *buf, char **start, off_t offset, int length, int *eof, void *data)
{
	struct netndis_private *priv = (struct netndis_private *)data;

	if(offset)
		return 0;

	if(length > PAGE_SIZE)
		length = PAGE_SIZE;

	snprintf(buf, length - 1, "%s\n", priv->pDevNode->hwInstName);
	buf[length-1] = '\0';
	return strlen(buf);
}

static int netndis_get_licstatus(char *buf, char **start, off_t offset, int length, int *eof, void *data)
{
	struct netndis_private *priv = (struct netndis_private *)data;

	if(offset)
		return 0;

	if(length > PAGE_SIZE)
		length = PAGE_SIZE;

	snprintf(buf, length - 1, "%s\n", priv->pDevNode->hwLicStatus);
	buf[length-1] = '\0';
	return strlen(buf);
}

static int netndis_get_hwaddr(char *buf, char **start, off_t offset, int length, int *eof, void *data)
{
	struct netndis_private *priv = (struct netndis_private *)data;

	if(offset)
		return 0;

	if(length > PAGE_SIZE)
		length = PAGE_SIZE;

	snprintf(buf, length - 1, "%02X:%02X:%02X:%02X:%02X:%02X\n",
			priv->pDevNode->hwAddr[0], priv->pDevNode->hwAddr[1],
			priv->pDevNode->hwAddr[2], priv->pDevNode->hwAddr[3],
			priv->pDevNode->hwAddr[4], priv->pDevNode->hwAddr[5]);
	buf[length-1] = '\0';
	return strlen(buf);
}

static int netndis_get_drivername(char *buf, char **start, off_t offset, int length, int *eof, void *data)
{
	struct netndis_private *priv = (struct netndis_private *)data;

	if(offset)
		return 0;

	if(length > PAGE_SIZE)
		length = PAGE_SIZE;

	snprintf(buf, length - 1, "%s\n", priv->pDevNode->hwDriverName ? priv->pDevNode->hwDriverName : "unknown");
	buf[length-1] = '\0';
	return strlen(buf);
}

static int
netndis_devinit(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err = 0;
	unsigned long val;
	int pid;

	TRACE_ENTER(dev->name);

	/* No need to lock, the hw_unavailable flag is already set in
	 * alloc_netndisdev() */

	priv->osthrd = OsThreadCreate(dev->name, FALSE, &pid);
	if(!priv->osthrd) {
		err = -ENOMEM;
		goto out;
	}

	if(pid <= 0) {
		OsThreadDestroy(priv->osthrd);
		priv->osthrd = NULL;
		if(pid < 0)
			err = pid;
		else
			err = -EIO;
		goto out;
	}

	err = NdisWrapStatusToErrno(
			NdisWrapInitializeAdapter(priv->miniportAdapter,
				netndis_status_handler,
				netndis_rx_handler,
				netndis_xmitcomplete_handler,
				netndis_resetcomplete_handler,
				netndis_hang_handler));
	if(err) {
		printk(KERN_ERR "%s: failed to initialize adapter (err = %d)\n",
		       dev->name, err);
		OsThreadDestroy(priv->osthrd);
		priv->osthrd = NULL;
		goto out;
	}

	err = NdisWrapStatusToErrno(
			NdisWrapGetMulticastListSize(priv->miniportAdapter, &priv->max_multicast));
	if(err)
		priv->max_multicast = 0;

#if 0
	priv->has_sensitivity = 0;
#endif

	priv->has_wep = priv->has_big_wep = NdisWrapStatusToErrno(NdisWrapGetEncode(priv->miniportAdapter, &val)) == 0;

	priv->has_pm = 0;

	priv->cfg = (1 << NETNDIS_CFG_LAST) - 1;

#if 0
	if (priv->has_wep) {
		printk(KERN_DEBUG "%s: WEP supported, ", dev->name);
		if (priv->has_big_wep)
			printk("104-bit key\n");
		else
			printk("40-bit key\n");
	}
#endif

	/* Get the MAC address */
	err = NdisWrapStatusToErrno(
			NdisWrapGetCurrentEthAddress(priv->miniportAdapter, dev->dev_addr, ETH_ALEN));
	if (err) {
		printk(KERN_WARNING "%s: failed to read MAC address!\n",
		       dev->name);
		NdisWrapHaltAdapter(priv->miniportAdapter);
		OsThreadDestroy(priv->osthrd);
		priv->osthrd = NULL;
		goto out;
	}

	/* Get the station name */
	memset(&priv->nick, 0, sizeof(priv->nick));
	strncpy(priv->nick, "unknown", sizeof(priv->nick));

	//printk(KERN_DEBUG "%s: Station name \"%s\"\n", dev->name, priv->nick);

#if 0
	/* Get allowed channels */
	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_CHANNELLIST,
				  &priv->channel_mask);
	if (err) {
		printk(KERN_ERR "%s: failed to read channel list!\n",
		       dev->name);
		NdisWrapHaltAdapter(priv->miniportAdapter);
		OsThreadDestroy(priv->osthrd);
		priv->osthrd = NULL;
		goto out;
	}
#else
	priv->channel_mask = 0xffff;
#endif

#if 0
	/* Get initial AP density */
	err = hhh_read_wordrec(hw, USER_BAP, HHH_RID_CNFSYSTEMSCALE,
				  &priv->ap_density);
	if (err || priv->ap_density < 1 || priv->ap_density > 3) {
		priv->has_sensitivity = 0;
	}
#endif

	/* Get initial RTS threshold */
	err = NdisWrapStatusToErrno(
			NdisWrapGetRtsThreshHold(priv->miniportAdapter, &val));
	if (err) {
		//printk(KERN_ERR "%s: failed to read RTS threshold!\n", dev->name);
		val = 2347;
	}
	priv->rts_thresh = val;

	/* Get initial fragmentation settings */
	err = NdisWrapStatusToErrno(
			NdisWrapGetFragThreshHold(priv->miniportAdapter, &val));
	if (err) {
		//printk(KERN_ERR "%s: failed to read fragmentation threshold!\n", dev->name);
		val = 2346;
		err = 0;
	}
	priv->frag_thresh = val;

	/* Power management setup */
	if (priv->has_pm) {
		ndiswrappowermode_t powermode;

		err = NdisWrapStatusToErrno(
				NdisWrapGetPowerMode(priv->miniportAdapter, &powermode));

		priv->pm_on = !err && (powermode > NDISWRAP_PWRMODE_CAM);
		priv->pm_mcast = 1;
		priv->pm_period = 0; /* XXX */
		priv->pm_timeout = 0; /* XXX */

		err = 0;
	}

	/* Set up the default configuration */
	priv->iw_mode = IW_MODE_INFRA;
	priv->channel = 10; /* default channel, more-or-less arbitrary */

	priv->promiscuous = 0;
	priv->iw_encode = IW_ENCODE_DISABLED;
	priv->tx_key = 0;

	/* Make the hardware available, as long as it hasn't been
	 * removed elsewhere (e.g. by PCMCIA hot unplug) */
	err = netndis_lock(priv, TRUE);
	if(err) {
		printk(KERN_ERR "%s: %s: Error %d locking\n", dev->name, __FUNCTION__, err);
	}
	priv->hw_unavailable--;
	if(!err)
		netndis_unlock(priv);
	else
		err = 0;

	if(netndis_proc_dir) {
		priv->proc_unit_dir = proc_mkdir(dev->name, netndis_proc_dir);
		if(priv->proc_unit_dir) {
			priv->proc_hwinst = create_proc_read_entry("hwinst", 0, priv->proc_unit_dir, netndis_get_hwinst, priv);
			priv->proc_hwaddr = create_proc_read_entry("hwaddr", 0, priv->proc_unit_dir, netndis_get_hwaddr, priv);
			priv->proc_drivername = create_proc_read_entry("drivername", 0, priv->proc_unit_dir, netndis_get_drivername, priv);
			priv->proc_licstatus = create_proc_read_entry("licstatus", 0, priv->proc_unit_dir, netndis_get_licstatus, priv);
		}
	}

 out:
	TRACE_EXIT(dev->name);
	return err;
}

static void
netndis_devuninit(struct net_device *dev)
{
	struct netndis_private *priv = dev->priv;
	int err, i;

	OsDebugPrintf("%s\n", __FUNCTION__);

	for(i=0; (i<5) && priv->work_scheduled; i++) {
		OsTestPrintf("%s: waiting.. irql=%s\n", __FUNCTION__, OsIrqlName(OsGetIrql()));
		OsSleep(100);
	}

	err = netndis_lock(priv, TRUE);
	if(err) {
		printk(KERN_ERR "%s: %s: Error %d locking\n", dev->name, __FUNCTION__, err);
	}
	priv->hw_unavailable++;
	if(!err)
		netndis_unlock(priv);
	else
		err = 0;

	for(i=0; (i<5) && priv->work_scheduled; i++) {
		OsTestPrintf("%s: waiting.. irql=%s\n", __FUNCTION__, OsIrqlName(OsGetIrql()));
		OsSleep(100);
	}

	NdisWrapHaltAdapter(priv->miniportAdapter);

	for(i=0; (i<5) && priv->work_scheduled; i++) {
		OsTestPrintf("%s: waiting.. irql=%s\n", __FUNCTION__, OsIrqlName(OsGetIrql()));
		OsSleep(100);
	}

	OsThreadDestroy(priv->osthrd);
	priv->osthrd = NULL;

	if(priv->proc_licstatus)
		remove_proc_entry("licstatus", priv->proc_unit_dir);
	if(priv->proc_drivername)
		remove_proc_entry("drivername", priv->proc_unit_dir);
	if(priv->proc_hwaddr)
		remove_proc_entry("hwaddr", priv->proc_unit_dir);
	if(priv->proc_hwinst)
		remove_proc_entry("hwinst", priv->proc_unit_dir);
	if(priv->proc_unit_dir)
		remove_proc_entry(dev->name, netndis_proc_dir);
}

struct net_device *alloc_netndisdev(POS_DEVNODE pDevNode)
{
	struct net_device *dev;
	struct netndis_private *priv;

	dev = alloc_etherdev(sizeof(struct netndis_private));
	if(!dev)
		return NULL;

	priv = (struct netndis_private *)dev->priv;
	memset(priv, 0, sizeof(struct netndis_private));
	priv->ndev = dev;
	memset(dev->name, 0, sizeof(dev->name));
	strncpy(dev->name, pDevNode->netDevName[0] ? pDevNode->netDevName : DLDRNETIF, sizeof(dev->name));

	/* Setup / override net_device fields */
	dev->init = netndis_devinit;
	dev->uninit = netndis_devuninit;
	dev->hard_start_xmit = netndis_xmit;
	dev->tx_timeout = netndis_tx_timeout;
	dev->watchdog_timeo = 4 * HZ; /* 4 second timeout */
	dev->get_stats = netndis_get_stats;
	dev->get_wireless_stats = netndis_get_wireless_stats;
	dev->do_ioctl = netndis_ioctl;
	dev->change_mtu = netndis_change_mtu;
	dev->set_multicast_list = netndis_set_multicast_list;
	/* we use the default eth_mac_addr for setting the MAC addr */

	/* Set up default callbacks */
	dev->open = netndis_open;
	dev->stop = netndis_stop;
	priv->pDevNode = pDevNode;
	priv->miniportAdapter = pDevNode->miniportAdapter;

	spin_lock_init(&priv->lock);
	init_waitqueue_head(&priv->wq);
	priv->mylock = 0;
	priv->open = 0;
	priv->hw_unavailable = 1; /* netndis_devinit() must clear this
				   * before anything else touches the
				   * hardware */
	//INIT_WORK(&priv->reset_work, (void (*)(void *))netndis_reset, dev);
	sema_init(&priv->reset_complete_sem, 0);

	//INIT_WORK(&priv->config_work, (void (*)(void *))netndis_config, dev);

	netif_carrier_off(dev);
	priv->last_linkstatus = NDISWRAP_LINKSTATUS_LAST;

	return dev;

}

void netndis_init(void)
{
	netndis_proc_dir = proc_mkdir(DLDRNAME, proc_root_driver);
}

void netndis_exit(void)
{
	if(netndis_proc_dir) {
		remove_proc_entry(DLDRNAME, proc_root_driver);
	}
}

